import os
from collections import OrderedDict
from pprint import pprint

import numpy as np
import pandas as pd

import sanpy

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

class bExport():
    """Once analysis is performed with sanpy.bAnalysis.spikeDetect(),
        reports can be generated with the bExport class.

    Example reports are:

    - Generating reports as a Pandas DataFrame.
    - (depreciated) Saving reports as a Microsoft Excel file.
    - Saving reports as a CSV text files.
    """
    def __init__(self, ba):
        """
        Args:
            ba (sanpy.bAnalysis): A bAnalysis object that has had spikes detected with detectSpikes().
        """
        self.ba = ba
        self.sweepNumber = 0

    def old_report(self, theMin, theMax):
        """
        Get entire spikeDict as a Pandas DataFrame.

        Args:
            theMin (float): Start seconds of the analysis
            theMax (float): Stop seconds of the analysis

        Returns:
            df: Pandas DataFrame
        """
        if theMin is None or theMax is None:
            #return None
            theMin = 0
            theMax = self.ba.fileLoader.recordingDur

        logger.info(f'theMin:{theMin} theMax:{theMax}')

        df = self.ba.asDataFrame()
        df = df[ (df['thresholdSec']>=theMin) & (df['thresholdSec']<=theMax) ]

        # added when trying to make scatterwidget for one file
        df['Condition'] = '' #df['condition1']
        df['File Number'] = '' #df['condition2']
        df['Sex'] = '' #df['condition3']
        df['Region'] = '' #df['condition4']

        # make new column with sex/region encoded
        '''
        tmpNewCol = 'RegSex'
        self.ba.masterDf[tmpNewCol] = ''
        for tmpRegion in ['Superior', 'Inferior']:
            for tmpSex in ['Male', 'Female']:
                newEncoding = tmpRegion[0] + tmpSex[0]
                regSex = self.ba.masterDf[ (self.ba.masterDf['Region']==tmpRegion) & (self.ba.masterDf['Sex']==tmpSex)]
                regSex = (self.ba.masterDf['Region']==tmpRegion) & (self.ba.masterDf['Sex']==tmpSex)
                print('newEncoding:', newEncoding, 'regSex:', regSex.shape)
                self.ba.masterDf.loc[regSex, tmpNewCol] = newEncoding
        '''

        # want this but region/sex/condition are not defined
        #print('bExport.report()')
        #print(df.head())
        tmpNewCol = 'CellTypeSex'
        cellTypeStr = df['cellType'].iloc[0]
        sexStr = df['sex'].iloc[0]
        #print('cellTypeStr:', cellTypeStr, 'sexStr:', sexStr)
        regSexEncoding = cellTypeStr + sexStr
        df[tmpNewCol] = regSexEncoding

        minStr = '%.2f'%(theMin)
        maxStr = '%.2f'%(theMax)
        minStr = minStr.replace('.', '_')
        maxStr = maxStr.replace('.', '_')

        # TODO: bytestreams are not strictly from a hdd folder or file
        fileName = self.ba.fileLoader.filename
        if fileName is not None:
            fileName, tmpExt = os.path.splitext(fileName)
            analysisName = fileName + '_s' + minStr + '_s' + maxStr
            #print('    minStr:', minStr, 'maxStr:', maxStr, 'analysisName:', analysisName)
        else:
            analysisName = 'bytestream'
        df['analysisname'] = analysisName

        return df

    def report2(self, theMin :float, theMax : float) -> pd.DataFrame:
        """
        Generate a human readable report of spikes.
        Include spike times between theMin and theMax (Sec).

        Args:
            theMin (float): Start seconds to save, inclusive
            theMax (float): Stop seconds to save, inclusive

        Returns:
            df: pd.DataFrame
        """
        newList = []
        for spikeIdx, spike in enumerate(self.ba.spikeDict):

            # if current spike time is out of bounds then continue (e.g. it is not between theMin (sec) and theMax (sec)
            spikeTime_sec = self.ba.fileLoader.pnt2Sec_(spike['thresholdPnt'])
            if spikeTime_sec<theMin or spikeTime_sec>theMax:
                continue

            spikeDict = OrderedDict() # use OrderedDict so Pandas output is in the correct order
            spikeDict['Spike'] = spikeIdx
            spikeDict['Take Off Potential (s)'] = self.ba.fileLoader.pnt2Sec_(spike['thresholdPnt'])
            spikeDict['Take Off Potential (ms)'] = self.ba.fileLoader.pnt2Ms_(spike['thresholdPnt'])
            spikeDict['Take Off Potential (mV)'] = spike['thresholdVal']
            spikeDict['AP Peak (ms)'] = self.ba.fileLoader.pnt2Ms_(spike['peakPnt'])
            spikeDict['AP Peak (mV)'] = spike['peakVal']
            spikeDict['AP Height (mV)'] = spike['peakHeight']
            spikeDict['Pre AP Min (mV)'] = spike['preMinVal']
            #spikeDict['Post AP Min (mV)'] = spike['postMinVal']
            #
            #spikeDict['AP Duration (ms)'] = spike['apDuration_ms']
            spikeDict['Early Diastolic Duration (ms)'] = spike['earlyDiastolicDuration_ms']
            spikeDict['Early Diastolic Depolarization Rate (dV/s)'] = spike['earlyDiastolicDurationRate'] # abb 202012
            spikeDict['Diastolic Duration (ms)'] = spike['diastolicDuration_ms']
            #
            spikeDict['Inter-Spike-Interval (ms)'] = spike['isi_ms']
            spikeDict['Spike Frequency (Hz)'] = spike['spikeFreq_hz']
            spikeDict['ISI (ms)'] = spike['isi_ms']

            spikeDict['Cycle Length (ms)'] = spike['cycleLength_ms']

            spikeDict['Max AP Upstroke (dV/dt)'] = spike['preSpike_dvdt_max_val2']
            spikeDict['Max AP Upstroke (mV)'] = spike['preSpike_dvdt_max_val']

            spikeDict['Max AP Repolarization (dV/dt)'] = spike['postSpike_dvdt_min_val2']
            spikeDict['Max AP Repolarization (mV)'] = spike['postSpike_dvdt_min_val']

            # half-width
            for widthDict in spike['widths']:
                keyName = 'width_' + str(widthDict['halfHeight'])
                spikeDict[keyName] = widthDict['widthMs']

            spikeDict['File'] = self.ba.fileLoader.filename

            # errors
            #spikeDict['numError'] = spike['numError']
            spikeDict['errors'] = spike['errors']


            # append
            newList.append(spikeDict)

        df = pd.DataFrame(newList)
        return df

    def getSummary(self, theMin : float = None, theMax : float = None) -> pd.DataFrame:
        """Get analysis summary as df.
        This adds some header information to spike report bExport.report2().
        """

        if self.ba.numSpikes == 0:
            logger.warning(f'did not find and spikes for summary')
            return None

        if theMin is None or theMax is None:
            theMin = 0
            theMax = self.ba.fileLoader.recordingDur

        #
        # cardiac style analysis to sheet 'cardiac'
        cardiac_df = self.report2(theMin, theMax) # report2 is more 'cardiac'

        dDict = self.ba.getDetectionDict()

        #
        # header sheet
        headerDict = OrderedDict()
        filePath, fileName = os.path.split(self.ba.fileLoader.filepath)
        headerDict['File Name'] = [fileName]
        headerDict['File Path'] = [filePath]

        headerDict['Cell Type'] = [dDict['cellType']]
        headerDict['Sex'] = [dDict['sex']]
        headerDict['Condition'] = [dDict['condition']]

        headerDict['Date Analyzed'] = [self.ba.analysisDate] # pulled from first detected spike
        headerDict['Time Analyzed'] = [self.ba.analysisTime]

        headerDict['Detection Type'] = [dDict['detectionType']]
        headerDict['dV/dt Threshold'] = [dDict['dvdtThreshold']]
        #headerDict['mV Threshold'] = [self.ba.mvThreshold] # abb 202012
        headerDict['Vm Threshold (mV)'] = [dDict['mvThreshold']]
        #headerDict['Median Filter (pnts)'] = [self.ba.medianFilter]
        headerDict['Analysis Version'] = [sanpy.analysisVersion]
        headerDict['Interface Version'] = [sanpy.interfaceVersion]

        #headerDict['Analysis Start (sec)'] = [self.ba.startSeconds]
        #headerDict['Analysis Stop (sec)'] = [self.ba.stopSeconds]
        headerDict['Sweep Number'] = ['Default 0']  # [self.ba.currentSweep]
        headerDict['Number of Sweeps'] = [self.ba.fileLoader.numSweeps]
        headerDict['Export Start (sec)'] = [float('%.2f'%(theMin))] # on export, x-axis of raw plot will be ouput
        headerDict['Export Stop (sec)'] = [float('%.2f'%(theMax))] # on export, x-axis of raw plot will be ouput

        # 'stats' has xxx columns (name, mean, sd, se, n)
        headerDict['stats'] = []

        ignoreColumns = ['Spike', 'File']
        for idx, col in enumerate(cardiac_df):
            if col in ignoreColumns:
                # in general, skip non numerical columns
                continue
            headerDict[col] = []

        # mean
        theMean = cardiac_df.mean(numeric_only=True) # skipna default is True
        theMean['errors'] = ''
        # sd
        theSD = cardiac_df.std(numeric_only=True) # skipna default is True
        theSD['errors'] = ''
        #se
        theSE = cardiac_df.sem(numeric_only=True) # skipna default is True
        theSE['errors'] = ''
        #n
        theN = cardiac_df.count(numeric_only=True) # skipna default is True
        theN['errors'] = ''

        statCols = ['mean', 'sd', 'se', 'n']
        for j, stat in enumerate(statCols):
            if j == 0:
                pass
            else:
                # need to append columns to keep Excel sheet columns in sync
                #for k,v in headerDict.items():
                #    headerDict[k].append('')

                headerDict['File Name'].append('')
                headerDict['File Path'].append('')
                headerDict['Cell Type'].append('')
                headerDict['Sex'].append('')
                headerDict['Condition'].append('')
                #
                headerDict['Date Analyzed'].append('')
                headerDict['Time Analyzed'].append('')
                headerDict['Detection Type'].append('')
                headerDict['dV/dt Threshold'].append('')
                headerDict['Vm Threshold (mV)'].append('')
                #headerDict['Median Filter (pnts)'].append('')
                headerDict['Analysis Version'].append('')
                headerDict['Interface Version'].append('')
                headerDict['Sweep Number'].append('')
                headerDict['Number of Sweeps'].append('')
                headerDict['Export Start (sec)'].append('')
                headerDict['Export Stop (sec)'].append('')

            # a dictionary key for each stat
            headerDict['stats'].append(stat)
            for idx, col in enumerate(cardiac_df):
                if col in ignoreColumns:
                    # in general, need to ignore string columns
                    #headerDict[col].append('')
                    continue
                #headerDict[col].append('')
                if stat == 'mean':
                    headerDict[col].append(theMean[col])
                elif stat == 'sd':
                    headerDict[col].append(theSD[col])
                elif stat == 'se':
                    headerDict[col].append(theSE[col])
                elif stat == 'n':
                    headerDict[col].append(theN[col])

        # end for j, stat
        #print('=== headerDict')
        #for k,v in headerDict.items():
        #    print(k, ':', v)

        # dict to pandas dataframe
        df = pd.DataFrame(headerDict).T
        df.insert(0, '', headerDict.keys(), allow_duplicates=True)

        return df

    def saveReport(self, savefile, theMin=None, theMax=None,
                    saveExcel=True, alsoSaveTxt=True, verbose=True):
        """
        Save a spike report for detected spikes between theMin (sec) and theMax (sec).

        This is used by main interface 'Export Spike Report'

        Args:
            savefile (str): path to xlsx file
            theMin (float): start/stop seconds of the analysis
            theMax (float): start/stop seconds of the analysis
            saveExcel (bool):
            alsoSaveTxt (bool):

        Return:
            str: analysisName
            df: df
        """
        if theMin is None or theMax is None:
            theMin = 0
            theMax = self.ba.fileLoader.recordingDur

        # always grab a df to the entire analysis (not sure what I will do with this)
        #df = self.ba.report() # report() is my own 'bob' verbiage

        theRet = None

        logger.warning('NEVER SAVING EXCEL !!! dec 2022')
        saveExcel = False
        if saveExcel and savefile:
            #if verbose: print('    bExport.saveReport() saving user specified .xlsx file:', savefile)
            excelFilePath = savefile
            writer = pd.ExcelWriter(excelFilePath, engine='xlsxwriter')

            #
            # cardiac style analysis to sheet 'cardiac'
            cardiac_df = self.report2(theMin, theMax) # report2 is more 'cardiac'

            dDict = self.ba.getDetectionDict()
            dateAnalyzed = self.ba.dateAnalyzed
            timeAnalyzed = self.ba.dateAnalyzed

            #
            # header sheet
            headerDict = OrderedDict()
            filePath, fileName = os.path.split(self.ba.filepath)
            headerDict['File Name'] = [fileName]
            headerDict['File Path'] = [filePath]

            headerDict['Cell Type'] = [dDict['cellType']]
            headerDict['Sex'] = [dDict['sex']]
            headerDict['Condition'] = [dDict['condition']]

            # todo: get these params in ONE dict inside self.ba
            #dateAnalyzed, timeAnalyzed = self.ba.dateAnalyzed.split(' ')
            headerDict['Date Analyzed'] = [dateAnalyzed]
            headerDict['Time Analyzed'] = [timeAnalyzed]
            headerDict['Detection Type'] = [dDict['detectionType']]
            headerDict['dV/dt Threshold'] = [dDict['dvdtThreshold']]
            #headerDict['mV Threshold'] = [self.ba.mvThreshold] # abb 202012
            headerDict['Vm Threshold (mV)'] = [dDict['mvThreshold']]
            #headerDict['Median Filter (pnts)'] = [self.ba.medianFilter]
            headerDict['Analysis Version'] = [sanpy.analysisVersion]
            headerDict['Interface Version'] = [sanpy.interfaceVersion]

            #headerDict['Analysis Start (sec)'] = [self.ba.startSeconds]
            #headerDict['Analysis Stop (sec)'] = [self.ba.stopSeconds]
            headerDict['Sweep Number'] = ['Default 0']  # [self.ba.currentSweep]
            headerDict['Number of Sweeps'] = [self.ba.fileLoader.numSweeps]
            headerDict['Export Start (sec)'] = [float('%.2f'%(theMin))] # on export, x-axis of raw plot will be ouput
            headerDict['Export Stop (sec)'] = [float('%.2f'%(theMax))] # on export, x-axis of raw plot will be ouput

            # 'stats' has xxx columns (name, mean, sd, se, n)
            headerDict['stats'] = []

            ignoreColumns = ['Spike', 'File']
            for idx, col in enumerate(cardiac_df):
                if col in ignoreColumns:
                    # in general, need to ignore string columns
                    #headerDict[col].append('')
                    continue
                headerDict[col] = []

            # mean
            theMean = cardiac_df.mean() # skipna default is True
            theMean['errors'] = ''
            # sd
            theSD = cardiac_df.std() # skipna default is True
            theSD['errors'] = ''
            #se
            theSE = cardiac_df.sem() # skipna default is True
            theSE['errors'] = ''
            #n
            theN = cardiac_df.count() # skipna default is True
            theN['errors'] = ''

            statCols = ['mean', 'sd', 'se', 'n']
            for j, stat in enumerate(statCols):
                if j == 0:
                    pass
                else:
                    # need to append columns to keep Excel sheet columns in sync
                    #for k,v in headerDict.items():
                    #    headerDict[k].append('')

                    headerDict['File Name'].append('')
                    headerDict['File Path'].append('')
                    headerDict['Cell Type'].append('')
                    headerDict['Sex'].append('')
                    headerDict['Condition'].append('')
                    #
                    headerDict['Date Analyzed'].append('')
                    headerDict['Time Analyzed'].append('')
                    headerDict['Detection Type'].append('')
                    headerDict['dV/dt Threshold'].append('')
                    headerDict['Vm Threshold (mV)'].append('')
                    #headerDict['Median Filter (pnts)'].append('')
                    headerDict['Analysis Version'].append('')
                    headerDict['Interface Version'].append('')
                    headerDict['Sweep Number'].append('')
                    headerDict['Number of Sweeps'].append('')
                    headerDict['Export Start (sec)'].append('')
                    headerDict['Export Stop (sec)'].append('')

                # a dictionary key for each stat
                headerDict['stats'].append(stat)
                for idx, col in enumerate(cardiac_df):
                    if col in ignoreColumns:
                        # in general, need to ignore string columns
                        #headerDict[col].append('')
                        continue
                    #headerDict[col].append('')
                    if stat == 'mean':
                        headerDict[col].append(theMean[col])
                    elif stat == 'sd':
                        headerDict[col].append(theSD[col])
                    elif stat == 'se':
                        headerDict[col].append(theSE[col])
                    elif stat == 'n':
                        headerDict[col].append(theN[col])

            #print(headerDict)
            #for k,v in headerDict.items():
            #    print(k, v)

            # dict to pandas dataframe
            df = pd.DataFrame(headerDict).T
            df.to_excel(writer, sheet_name='summary')

            # set the column widths in excel sheet 'cardiac'
            columnWidth = 25
            worksheet = writer.sheets['summary']  # pull worksheet object
            for idx, col in enumerate(df):  # loop through all columns
                worksheet.set_column(idx, idx, columnWidth)  # set column width

            #
            # 'params' sheet with all detection params
            # need to convert list values in dict to string (o.w. we get one row per item in list)
            exportDetectionDict = {}
            for k, v in dDict.items():
                # v is a dict from bDetection
                if isinstance(v, list):
                    v = f'"{v}"'
                exportDetectionDict[k] = v
            #print('  === "params" sheet exportDetectionDict:', exportDetectionDict)
            #df = pd.DataFrame(exportDetectionDict, index=[0]).T # index=[0] needed when dict has all scalar values
            detection_df = pd.DataFrame(exportDetectionDict).T
            detection_df.to_excel(writer, sheet_name='params')
            # worksheet is <class 'xlsxwriter.worksheet.Worksheet'>
            worksheet = writer.sheets['params']  # pull worksheet object
            # set first 20 columns to columnWidth
            columnWidth = 18
            startCol = 0
            stopCol = 20 # xlswriter.worksheet does not care about the stop column
            worksheet.set_column(0, stopCol, columnWidth)  # set column width

            #
            # 'cardiac' sheet with human readable stat names
            cardiac_df.to_excel(writer, sheet_name='cardiac')

            # set the column widths in excel sheet 'cardiac'
            columnWidth = 20
            worksheet = writer.sheets['cardiac']  # pull worksheet object
            for idx, col in enumerate(cardiac_df):  # loop through all columns
                worksheet.set_column(idx, idx, columnWidth)  # set column width


            #
            # mean spike clip
            theseClips, theseClips_x, meanClip = self.ba.getSpikeClips(theMin, theMax, sweepNumber=self.sweepNumber)
            try:
                first_X = theseClips_x[0] #- theseClips_x[0][0]
                #if verbose: print('    bExport.saveReport() saving mean clip to sheet "Avg Spike" from', len(theseClips), 'clips')
                df = pd.DataFrame(meanClip, first_X)
                df.to_excel(writer, sheet_name='Avg Spike')
            except (IndexError) as e:
                logger.warning('Got bad spike clips. Usually happend when 1-2 spikes')

            writer.save()

        #
        # save a csv text file
        #
        analysisName = ''
        if alsoSaveTxt:
            # this also saves
            analysisName, df0 = self.getReportDf(theMin, theMax, savefile)

            #
            # save mean spike clip

            not_used_theseClips, theseClips_x, meanClip = self.ba.getSpikeClips(theMin, theMax, sweepNumber=self.sweepNumber)
            if len(theseClips_x) == 0:
                pass
            else:
                first_X = theseClips_x[0] #- theseClips_x[0][0]
                first_X = np.array(first_X)
                first_X /= self.ba.fileLoader.dataPointsPerMs # pnts to ms
                #if verbose: print('    bExport.saveReport() saving mean clip to sheet "Avg Spike" from', len(theseClips), 'clips')
                #dfClip = pd.DataFrame(meanClip, first_X)
                dfClip = pd.DataFrame.from_dict({
                    'xMs': first_X,
                    'yVm': meanClip
                    })
                # load clip based on analysisname (with start/stop seconds)
                analysisname = df0['analysisname'].iloc[0] # name with start/stop seconds
                logger.info(f'analysisname: {analysisname}')
                clipFileName = analysisname + '_clip.csv'
                tmpPath, tmpFile = os.path.split(savefile)
                tmpPath = os.path.join(tmpPath, 'analysis')
                # dir is already created in getReportDf
                if not os.path.isdir(tmpPath):
                    os.mkdir(tmpPath)
                clipSavePath = os.path.join(tmpPath, clipFileName)
                logger.info(f'clipSavePath: {clipSavePath}')
                dfClip.to_csv(clipSavePath)
            #
            theRet = df0
        #
        return analysisName, theRet

    def getReportDf(self, theMin, theMax, savefile):
        """Get spikes as a Pandas DataFrame, one row per spike.

        Args:
            theMin (float): xxx
            theMax (float): xxx
            savefile (str): .xls file path

        Returns:
            df: Pandas DataFrame
        """
        filePath, fileName = os.path.split(os.path.abspath(savefile))

        dDict = self.ba.getDetectionDict()

        # make an analysis folder
        filePath = os.path.join(filePath, 'analysis')
        if not os.path.isdir(filePath):
            logger.info(f'Making output folder: {filePath}')
            os.mkdir(filePath)

        textFileBaseName, tmpExtension = os.path.splitext(fileName)
        textFilePath = os.path.join(filePath, textFileBaseName + '.csv')

        # save header
        textFileHeader = OrderedDict()
        textFileHeader['file'] = self.ba.fileLoader.filename
        #textFileHeader['condition1'] = self.ba.condition1
        #textFileHeader['condition2'] = self.ba.condition2
        #textFileHeader['condition3'] = self.ba.condition3
        textFileHeader['cellType'] = dDict['cellType']
        textFileHeader['sex'] = dDict['sex']
        textFileHeader['condition'] = dDict['condition']
        #
        textFileHeader['dateAnalyzed'] = self.ba.dateAnalyzed
        textFileHeader['detectionType'] = dDict['detectionType']
        textFileHeader['dvdtThreshold'] = [dDict['dvdtThreshold']]
        textFileHeader['mvThreshold'] = [dDict['mvThreshold']]
        #textFileHeader['medianFilter'] = self.ba.medianFilter
        textFileHeader['startSeconds'] = '%.2f'%(theMin)
        textFileHeader['stopSeconds'] = '%.2f'%(theMax)
        #textFileHeader['startSeconds'] = self.ba.startSeconds
        #textFileHeader['stopSeconds'] = self.ba.stopSeconds
        textFileHeader['currentSweep'] = 'Default 0' # self.ba.currentSweep
        textFileHeader['numSweeps'] = self.ba.fileLoader.numSweeps
        #textFileHeader['theMin'] = theMin
        #textFileHeader['theMax'] = theMax

        # 20210125, this is not needed, we are saviing pandas df below ???
        headerStr = ''
        for k,v in textFileHeader.items():
            headerStr += k + '=' + str(v) + ';'
        headerStr += '\n'
        #print('headerStr:', headerStr)
        with open(textFilePath,'w') as f:
            f.write(headerStr)

        #df = self.report(theMin, theMax)
        df = self.ba.asDataFrame()
        
        # we need a column indicating (path), the original .abf file
        # along with (start,stop) which should make this analysis unique?
        minStr = '%.2f'%(theMin)
        maxStr = '%.2f'%(theMax)
        minStr = minStr.replace('.', '_')
        maxStr = maxStr.replace('.', '_')
        tmpPath, tmpFile = os.path.split(self.ba.fileLoader.filepath)
        tmpFile, tmpExt = os.path.splitext(tmpFile)
        analysisName = tmpFile + '_s' + minStr + '_s' + maxStr
        logger.info(f'minStr:{minStr} maxStr:{maxStr} analysisName:{analysisName}')
        df['analysisname'] = analysisName

        # should be filled in by self.ba.report
        #df['Condition'] =     df['condition1']
        #df['File Number'] =     df['condition2']
        #df['Sex'] =     df['condition3']
        #df['Region'] =     df['condition4']
        df['filename'] = [os.path.splitext(os.path.split(x)[1])[0] for x in     df['file'].tolist()]

        #
        logger.info('saving text file: {textFilePath}')
        #df.to_csv(textFilePath, sep=',', index_label='index', mode='a')
        df.to_csv(textFilePath, sep=',', index_label='index', mode='w')

        return analysisName, df

def test():
    path = 'data/19114001.abf'
    ba = sanpy.bAnalysis(path)

    bd = sanpy.bDetection()  # gets default
    dDict = bd.getDetectionDict('SA Node')
    ba.spikeDetect(dDict)

    be = bExport(ba)
    df = be.getSummary()
    logger.info('')
    print(df)

if __name__ == '__main__':
    test()



import sys
from typing import Union, Dict, List, Tuple, Optional

import numpy as np
import scipy.signal
import seaborn as sns
import matplotlib.pyplot as plt

import sanpy

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

def old_getEddLines(ba):
    """Get lines representing linear fit of EDD rate.

    Args:
        ba (bAnalysis): bAnalysis object
    """
    logger.info(ba)

    x = []
    y = []
    if ba is None or ba.numSpikes == 0:
        return x, y

    # these are getting for current sweep
    preLinearFitPnt0 = ba.getStat('preLinearFitPnt0')
    preLinearFitSec0 = [ba.pnt2Sec_(x) for x in preLinearFitPnt0]
    preLinearFitVal0 = ba.getStat('preLinearFitVal0')

    preLinearFitPnt1 = ba.getStat('preLinearFitPnt1')
    preLinearFitSec1 = [ba.pnt2Sec_(x) for x in preLinearFitPnt1]
    preLinearFitVal1 = ba.getStat('preLinearFitVal1')

    thisNumSpikes = len(preLinearFitPnt0)
    #for idx, spike in enumerate(range(ba.numSpikes)):
    for idx, spike in enumerate(range(thisNumSpikes)):
        dx = preLinearFitSec1[idx] - preLinearFitSec0[idx]
        dy = preLinearFitVal1[idx] - preLinearFitVal0[idx]

        lineLength = 8  # TODO: make this a function of spike frequency?

        x.append(preLinearFitSec0[idx])
        x.append(preLinearFitSec1[idx] + lineLength*dx)
        x.append(np.nan)

        y.append(preLinearFitVal0[idx])
        y.append(preLinearFitVal1[idx] + lineLength*dy)
        y.append(np.nan)

        #print('here')

    return x, y

def old_getHalfWidths(ba):
    """Get lines representing half-widhts (AP Dur).

    Args:
        ba (bAnalysis): bAnalysis object

    Returns:
        x,y (enum):
    """
    # defer until we know how many half-widths 20/50/80
    x = []
    y = []
    numPerSpike = 3  # rise/fall/nan
    #numSpikes = ba.numSpikes
    xyIdx = 0
    spikeDictionaries = ba.getSpikeDictionaries() # for current sweep
    numSpikes = len(spikeDictionaries)
    #for idx, spike in enumerate(ba.spikeDict):
    for idx, spike in enumerate(spikeDictionaries):
        if idx ==0:
            # make x/y from first spike using halfHeights = [20,50,80]
            halfHeights = spike['halfHeights'] # will be same for all spike, like [20, 50, 80]
            numHalfHeights = len(halfHeights)
            # *numHalfHeights to account for rise/fall + padding nan
            x = [np.nan] * (numSpikes * numHalfHeights * numPerSpike)
            y = [np.nan] * (numSpikes * numHalfHeights * numPerSpike)
            #print('  len(x):', len(x), 'numHalfHeights:', numHalfHeights, 'numSpikes:', numSpikes, 'halfHeights:', halfHeights)

        for idx2, width in enumerate(spike['widths']):
            halfHeight = width['halfHeight'] # [20,50,80]
            risingPnt = width['risingPnt']
            risingVal = width['risingVal']
            fallingPnt = width['fallingPnt']
            fallingVal = width['fallingVal']

            if risingPnt is None or fallingPnt is None:
                # half-height was not detected
                continue

            risingSec = ba.pnt2Sec_(risingPnt)
            fallingSec = ba.pnt2Sec_(fallingPnt)

            x[xyIdx] = risingSec
            x[xyIdx+1] = fallingSec
            x[xyIdx+2] = np.nan
            # y
            y[xyIdx] = fallingVal  #risingVal, to make line horizontal
            y[xyIdx+1] = fallingVal
            y[xyIdx+2] = np.nan

            # each spike has 3x pnts: rise/fall/nan
            xyIdx += numPerSpike  # accounts for rising/falling/nan
        # end for width
    # end for spike
    return x, y

class bAnalysisPlot():
    """
    Class to plot results of [sanpy.bAnalysis][sanpy.bAnalysis] spike detection.
    
    """
    def __init__(self, ba=None):
        """
        Args:
            ba: [sanpy.bAnalysis][sanpy.bAnalysis] object
        """
        self._ba = ba

    @property
    def ba(self):
        """Get underlying bAnalysis object."""
        return self._ba

    def getDefaultPlotStyle(self):
        """Get dictionary with default plot style."""
        d = {
            'linewidth': 1,
            'color': 'k',
            'width': 9,
            'height': 3,
        }
        return d.copy()

    def _makeFig(self, plotStyle=None):
        if plotStyle is None:
            plotStyle = self.getDefaultPlotStyle()

        grid = plt.GridSpec(1, 1, wspace=0.2, hspace=0.4)

        width = plotStyle['width']
        height = plotStyle['height']
        fig = plt.figure(figsize=(width, height))
        ax = fig.add_subplot(grid[0, 0:]) #Vm, entire sweep

        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)

        return fig, ax

    def plotRaw(self, plotStyle=None, ax=None):
        """
        Plot raw recording

        Args:
            plotStye (float):
            ax (xxx):
        """

        if plotStyle is None:
            plotStyle = self.getDefaultPlotStyle()

        if ax is None:
            _fig, _ax = self._makeFig()
            _fig.suptitle(f'{self._ba.fileLoader.filepath}')
        else:
            _ax = ax

        color = plotStyle['color']
        linewidth = plotStyle['linewidth']
        sweepX = self.ba.fileLoader.sweepX
        sweepY = self.ba.fileLoader.sweepY

        _ax.plot(sweepX, sweepY, '-', c=color, linewidth=linewidth) # fmt = '[marker][line][color]'

        xUnits = self.ba.fileLoader.get_xUnits()
        yUnits = self.ba.fileLoader.get_yUnits()
        _ax.set_xlabel(xUnits)
        _ax.set_ylabel(yUnits)

        return _ax
        
    def plotDerivAndRaw(self):
        """
        Plot both Vm and the derivative of Vm (dV/dt).

        Args:
            fig (matplotlib.pyplot.figure): An existing figure to plot to.
                see: https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.figure.html

        Return:
            fig and axs
        """

        #
        # make a 2-panel figure
        grid = plt.GridSpec(2, 1, wspace=0.2, hspace=0.4)
        fig = plt.figure(figsize=(10, 8))
        ax1 = fig.add_subplot(grid[0,0])
        ax2 = fig.add_subplot(grid[1,0], sharex=ax1)
        ax1.spines['right'].set_visible(False)
        ax1.spines['top'].set_visible(False)
        ax2.spines['right'].set_visible(False)
        ax2.spines['top'].set_visible(False)

        self.plotRaw(ax=ax1);

        sweepX = self.ba.fileLoader.sweepX
        filteredDeriv = self.ba.fileLoader.filteredDeriv
        ax2.plot(sweepX, filteredDeriv)

        ax2.set_ylabel('dV/dt')
        #ax2.set_xlabel('Seconds')

        return fig, [ax1, ax2]

    def plotSpikes(self,
                   plotThreshold=False, plotPeak=False,
                   plotStyle=None,
                   hue : Optional[str] = 'condition',
                   markerSize : Optional[int] = 4,
                   ax=None):
        '''Plot Vm with spike analysis overlaid as symbols

        Args:
            plotThreshold
            plotPeak
            plotStyle (dict): xxx
            markerSize
            ax (xxx): If specified will plot into a MatPlotLib axes

        Returns
            fig
            ax
        '''

        legend = False
        
        if plotStyle is None:
            plotStyle = self.getDefaultPlotStyle()

        if ax is None:
            fig, ax = self._makeFig()
            fig.suptitle(self.ba.getFileName())
        # plot vm
        self.plotRaw(ax=ax)

        # plot spike times
        if plotThreshold:
            thresholdSec = self.ba.getStat('thresholdSec')  # x
            thresholdVal = self.ba.getStat('thresholdVal')  # y
            
            ax.plot(thresholdSec, thresholdVal, 'pr', markersize=markerSize)
            df = self.ba.spikeDict.asDataFrame()  # regenerates, can be expensive
            #logger.info(f'df hue unique is: {df[hue].unique()}')
            #sns.scatterplot(x='thresholdSec', y='thresholdVal', hue=hue, data=df, ax=ax, legend=legend)

        # plot the peak
        if plotPeak:
            peakPnt = self.ba.getStat('peakPnt')
            # don't use this for Ca++ concentration
            #peakVal = self.ba.getStat('peakVal')
            peakVal = self.ba.fileLoader.sweepY[peakPnt]
            peakSec = [self.ba.fileLoader.pnt2Sec_(x) for x in peakPnt]
            ax.plot(peakSec, peakVal, 'oc', markersize=markerSize)

        xUnits = self.ba.fileLoader.get_xUnits()
        yUnits = self.ba.fileLoader.get_yUnits()
        ax.set_xlabel(xUnits)
        ax.set_ylabel(yUnits)

        return ax
    
    def plotStat(self, xStat : str, yStat : str,
                 hue : Optional[str] = None,
                 ax = None):
                 # ax : Optional["matplotlib.axes._subplots.AxesSubplot"] = None):
        
        legend = False
        
        if ax is None:
            fig, ax = self._makeFig()
            fig.suptitle(self.ba.getFileName())
        print(type(ax))
              
        df = self.ba.spikeDict.asDataFrame()  # regenerates, can be expensive
        sns.scatterplot(x=xStat, y=yStat, hue=hue, data=df, ax=ax, legend=legend)

    def plotTimeSeries(ba, stat, halfWidthIdx=0, ax=None):
        """ Plot a given spike parameter.
        """
        if stat == 'peak':
            yStatName = 'peakVal'
            yStatLabel = 'Spike Peak (mV)'
        if stat == 'preMin':
            yStatName = 'preMinVal'
            yStatLabel = 'Pre Min (mV)'
        if stat == 'halfWidth':
            yStatName = 'widthPnts'
            yStatLabel = 'Spike Half Width (ms)'

        #
        # pull
        statX = []
        statVal = []
        for i, spike in enumerate(ba.spikeDict):
            if i==0 or i==len(ba.spikeTimes)-1:
                continue
            else:
                statX.append(spike['peakSec'])
                if stat == 'halfWidth':
                    statVal.append(spike['widths'][halfWidthIdx]['widthMs'])
                else:
                    statVal.append(spike[yStatName])

        #
        # plot
        if ax is None:
            grid = plt.GridSpec(1, 1, wspace=0.2, hspace=0.4)

            fig = plt.figure(figsize=(10, 8))
            ax = fig.add_subplot(grid[0, 0:]) #Vm, entire sweep

        ax.plot(statX, statVal, 'o-k')

        ax.set_ylabel(yStatLabel)
        ax.set_xlabel('Time (sec)')

        return statVal

    def plotISI(ba, ax=None):
        """ Plot the inter-spike-interval (sec) between each spike threshold"""
        #
        # pull
        spikeTimes_sec = [x/ba.dataPointsPerMs/1000 for x in ba.spikeTimes]
        isi = np.diff(spikeTimes_sec)
        isi_x = spikeTimes_sec[0:-1]

        #
        # plot
        if ax is None:
            grid = plt.GridSpec(1, 1, wspace=0.2, hspace=0.4)

            fig = plt.figure(figsize=(10, 8))
            ax = fig.add_subplot(grid[0, 0:]) #Vm, entire sweep

        ax.plot(isi_x, isi, 'o-k')

        ax.set_ylabel('Inter-Spike-Interval (sec)')
        ax.set_xlabel('Time (sec)')

    def plotClips(self, plotType='Raw', preClipWidth_ms=None, postClipWidth_ms=None, ax=None):
        """Plot clips of all detected spikes

        Clips are created in self.spikeDetect() and default to clipWidth_ms = 100 ms

        Args:
            plotType (str): From [Raw, RawPlusMean, Mean, SD, Var]
            #plotVariance (bool): If tru, plot variance. Otherwise plot all raw clips (black) with mean (red)

        Returns:
            xPlot
            yPlot
        """

        '''
        if ax is None:
            grid = plt.GridSpec(1, 1, wspace=0.2, hspace=0.4)

            fig = plt.figure(figsize=(10, 8))
            ax = fig.add_subplot(grid[0, 0:]) #Vm, entire sweep
        '''

        if self.ba.numSpikes == 0:
            return None, None

        if ax is None:
            fig, ax = self._makeFig()
            fig.suptitle(self.ba.getFileName())

        startSec, stopSec = None, None
        selectedSpikeList = []
        #preClipWidth_ms = 200
        #postClipWidth_ms = 1000

        # Leave this none, if we pass none, ba.getSpikeClipsWill take care of this
        '''
        if preClipWidth_ms is None:
            preClipWidth_ms = self.ba.detectionClass['preSpikeClipWidth_ms']
        if postClipWidth_ms is None:
            postClipWidth_ms = self.ba.detectionClass['postSpikeClipWidth_ms']
        '''
        sweepNumber = 0
        theseClips, theseClips_x, meanClip = self.ba.getSpikeClips(startSec, stopSec,
                                                spikeSelection=selectedSpikeList,
                                                preSpikeClipWidth_ms=preClipWidth_ms,
                                                postSpikeClipWidth_ms=postClipWidth_ms,
                                                sweepNumber=sweepNumber)
        numClips = len(theseClips)

        # convert clips to 2d ndarray ???
        xTmp = np.array(theseClips_x)
        #xTmp /= self.ba.dataPointsPerMs * 1000  # pnt to seconds
        xTmp /= 1000  # ms to seconds
        yTmp = np.array(theseClips)  # mV

        # TODO: plot each clip as different color
        # this will show variation in sequential clips (for Ca++ imaging they are decreasing)
        #print(xTmp.shape)  # ( e.g. (9,306)
        #sys.exit(1)

        # plot variance
        if plotType == 'Mean':
            xPlot = np.nanmean(xTmp, axis=0) # xTmp is in ms
            yPlot = np.nanmean(yTmp, axis=0)
            ax.plot(xPlot, yPlot, '-k', linewidth=1)
            ax.set_ylabel('Mean')
            ax.set_xlabel('Time (sec)')
        elif plotType == 'Var':
            xPlot = np.nanmean(xTmp, axis=0) # xTmp is in ms
            yPlot = np.nanvar(yTmp, axis=0)
            ax.plot(xPlot, yPlot, '-k', linewidth=1)
            ax.set_ylabel('Variance')
            ax.set_xlabel('Time (sec)')
        elif plotType == 'SD':
            xPlot = np.nanmean(xTmp, axis=0) # xTmp is in ms
            yPlot = np.nanstd(yTmp, axis=0)
            ax.plot(xPlot, yPlot, '-k', linewidth=1)
            ax.set_ylabel('STD')
            ax.set_xlabel('Time (sec)')
        elif plotType in ['Raw', 'RawPlusMean']:
            # plot raw
            #logger.info('PLOTTING RAW')

            cmap = plt.get_cmap('jet')
            colors = [cmap(i) for i in np.linspace(0, 1, numClips)]

            for i in range(numClips):
                color = colors[i]
                xPlot = xTmp[i,:]
                yPlot = yTmp[i,:]
                # I want different colors here
                #ax.plot(xPlot, yPlot, '-k', linewidth=0.5)
                # WHY IS THIS NOT WORKING !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                #logger.info(f'!!!!!!!!!!!!!!!!!!!!!! plotting color g {plotType} {color}')
                #ax.plot(xPlot, yPlot, '-g', linewidth=0.5, color='g')
                ax.plot(xPlot, yPlot, '-', label=f'{i}', color=color, linewidth=0.5)

            yLabel = self.ba._sweepLabelY
            ax.set_ylabel(yLabel)
            ax.set_xlabel('Time (sec)')

            # plot mean
            if plotType == 'RawPlusMean':
                xMeanClip = np.nanmean(xTmp, axis=0) # xTmp is in ms
                yMeanClip = np.nanmean(yTmp, axis=0)
                ax.plot(xMeanClip, yMeanClip, '-r', linewidth=1)

            # show legend for each raw race 0,1,2,3,...
            ax.legend(loc='best')

        else:
            logger.error(f'Did not understand plot type: {plotType}')

        #
        return xPlot, yPlot

    def plotPhasePlot(self, oneSpikeNumber=None, ax=None):
        if ax is None:
            grid = plt.GridSpec(1, 1, wspace=0.2, hspace=0.4)

            fig = plt.figure(figsize=(10, 8))
            ax = fig.add_subplot(grid[0, 0:]) #Vm, entire sweep

        filteredClip = scipy.signal.medfilt(self.spikeClips[oneSpikeNumber],3)
        dvdt = np.diff(filteredClip)
        # add an initial point so it is the same length as raw data in abf.sweepY
        dvdt = np.concatenate(([0],dvdt))
        line, = ax.plot(filteredClip, dvdt, 'y')

        ax.set_ylabel('filtered dV/dt')
        ax.set_xlabel('filtered Vm (mV)')

        return line

def test_plot(path):
    print('=== test_plot() path:', path)
    ba = sanpy.bAnalysis(path)

    # detect
    dDict = sanpy.bAnalysis.getDefaultDetection()
    dDict['dvdThreshold'] = 50
    ba.spikeDetect(dDict)

    # plot
    bp = sanpy.bAnalysisPlot(ba)

    fig = bp.plotDerivAndRaw()

    fig = bp.plotSpikes()

    plt.show()

if __name__ == '__main__':
    path = 'data/19114001.abf'
    test_plot(path)

    # TODO: check if error

    path = 'data/19114001.csv'
    test_plot(path)



"""
Test bAnalysis.py

Run this file with

```
python -m unittest tests/test_analysis.py
```
"""

import os, shutil, tempfile
import unittest

import sanpy

import logging
from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__, level=logging.DEBUG)

class Test_Analysis_Dir(unittest.TestCase):
    def setUp(self):
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()

        # copy data into it
        src = os.path.join('data', '19114001.abf')
        dst = os.path.join(self.test_dir, '19114001.abf')
        #shutil.copyfile('data/19114001.abf', dst)
        shutil.copyfile(src, dst)

        src = os.path.join('data', '19114000.abf')
        dst = os.path.join(self.test_dir, '19114000.abf')
        #shutil.copyfile('data/19114000.abf', dst)
        shutil.copyfile(src, dst)

    def tearDown(self):
        # Remove the directory after the test
        shutil.rmtree(self.test_dir)

    def testLoadDir(self):

        logger.info('')
        
        ad = sanpy.analysisDir(path=self.test_dir)

        self.assertEqual(len(ad._df), 2)

        ad.saveHdf()

        ad.loadHdf()
        self.assertEqual(len(ad._df), 2)

        # add a new file and test sync folder
        src = os.path.join('data', '20191009_0005.abf')
        dst = os.path.join(self.test_dir, '20191009_0005.abf')
        shutil.copyfile(src, dst)

        logger.info(f'testing syncDfWithPath')
        ad.syncDfWithPath()
        self.assertEqual(len(ad._df), 3)

        # get all three files
        ba0 = ad.getAnalysis(0)
        ba1 = ad.getAnalysis(1)
        ba2 = ad.getAnalysis(2)

        # delete a row
        #ad.deleteRow(1)

        # remove from database
        #ad.removeRowFromDatabase(0)

if __name__ == '__main__':
    unittest.main()



"""
The bDetection class provides an interface to get and set detection paramers.

Example
-------

```python
import sanpy

# grab the presets for 'SA Node' cells
dDict = sanpy.bDetection().getDetectionDict('SA Node')
ba.spikeDetect(dDict)

# tweek individual parameters
dDict['dvdtThreshold'] = 50

# load a recording
myPath = 'data/19114001.abf'
ba = sanpy.bAnalysis(myPath)

# perform spike detection
ba.spikeDetect(dDict)

# browse results
```
"""
from http.client import RemoteDisconnected
import os
import numbers
import math
from enum import Enum
import json
import pathlib
from pprint import pprint
import glob
import copy
from collections import OrderedDict

from matplotlib.font_manager import json_load
#from colin.stochAnalysis import load

import sanpy

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

class detectionTypes_(Enum):
    """
    Detection type is one of (dvdt, mv).

    dvdt: Search for threshold crossings in first derivative of membrane potential.
    mv: Search for threshold crossings in membrane potential.
    """
    dvdt = 'dvdt'
    mv = 'mv'

# TODO (Cudmore) this needs to be a class so we can expand/contract based
# on what we find on the hard-drive
# allow user to save to detection presets

def getDefaultDetection():
    """Get detection parameters
    
    This includes a mapping from backend variable names to
    front-end human readable and long-format descriptions.

    Args:
        detectionPreset (enum): bDetection.detectionPresets.default

    Returns:
        dict: The default detection dictionary.
    """
    
    theDict = OrderedDict()  # {}

    '''
    key = 'include'
    theDict[key] = {}
    theDict[key]['defaultValue'] = True
    theDict[key]['type'] = 'bool'
    theDict[key]['allowNone'] = False
    theDict[key]['units'] = ''
    theDict[key]['humanName'] = 'Include'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'Include analysis for this file'
    '''

    key = 'detectionName'
    theDict[key] = {}
    theDict[key]['defaultValue'] = 'default' #detectionPreset.value # ('dvdt', 'mv')
    theDict[key]['type'] = 'string'
    theDict[key]['allowNone'] = False  # To do, have 2x entry points to bAnalysis detect, never set this to nan
    theDict[key]['units'] = ''
    theDict[key]['humanName'] = 'Detection Preset Name'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'The name of detection preset'

    key = 'userSaveName'
    theDict[key] = {}
    theDict[key]['defaultValue'] = '' #detectionPreset.value # ('dvdt', 'mv')
    theDict[key]['type'] = 'string'
    theDict[key]['allowNone'] = False  # To do, have 2x entry points to bAnalysis detect, never set this to nan
    theDict[key]['units'] = ''
    theDict[key]['humanName'] = 'Saved Detection Params'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'The name of saved user detection params'

    key = 'detectionType'
    theDict[key] = {}
    theDict[key]['defaultValue'] = sanpy.bDetection.detectionTypes['dvdt'].value  # ('dvdt', 'mv')
    theDict[key]['type'] = 'sanpy.bDetection.detectionTypes'
    theDict[key]['allowNone'] = False  # To do, have 2x entry points to bAnalysis detect, never set this to nan
    theDict[key]['units'] = ''
    theDict[key]['humanName'] = 'Detection Type'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'Detect using derivative (dvdt) or membrane potential (mV)'

    key = 'dvdtThreshold'
    theDict[key] = {}
    theDict[key]['defaultValue'] = 20
    theDict[key]['type'] = 'float'
    theDict[key]['allowNone'] = True  # To do, have 2x entry points to bAnalysis detect, never set this to nan
    theDict[key]['units'] = 'dVdt'
    theDict[key]['humanName'] = 'dV/dt Threshold'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'dV/dt threshold for a spike, will be backed up to dvdt_percentOfMax and have xxx error when this fails'

    key = 'mvThreshold'
    theDict[key] = {}
    theDict[key]['defaultValue'] = -20
    theDict[key]['type'] = 'float'
    theDict[key]['allowNone'] = False
    theDict[key]['units'] = 'mV'
    theDict[key]['humanName'] = 'mV Threshold'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'mV threshold for spike AND minimum spike mV when detecting with dV/dt'

    key = 'startSeconds'
    theDict[key] = {}
    theDict[key]['defaultValue'] = None
    theDict[key]['type'] = 'float'
    theDict[key]['allowNone'] = True
    theDict[key]['units'] = 's'
    theDict[key]['humanName'] = 'Start(s)'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'Start seconds of analysis'

    key = 'stopSeconds'
    theDict[key] = {}
    theDict[key]['defaultValue'] = None
    theDict[key]['type'] = 'float'
    theDict[key]['allowNone'] = True
    theDict[key]['units'] = 's'
    theDict[key]['humanName'] = 'Stop(s)'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'Stop seconds of analysis'

    key = 'cellType'
    theDict[key] = {}
    theDict[key]['defaultValue'] = ''
    theDict[key]['type'] = 'string'
    theDict[key]['allowNone'] = False
    theDict[key]['units'] = ''
    theDict[key]['humanName'] = 'Cell Type'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'Cell Type'

    key = 'sex'
    theDict[key] = {}
    theDict[key]['defaultValue'] = ''
    theDict[key]['type'] = 'string'
    theDict[key]['allowNone'] = False
    theDict[key]['units'] = ''
    theDict[key]['humanName'] = 'Sex'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'Sex'

    key = 'condition'
    theDict[key] = {}
    theDict[key]['defaultValue'] = ''
    theDict[key]['type'] = 'string'
    theDict[key]['allowNone'] = False
    theDict[key]['units'] = ''
    theDict[key]['humanName'] = 'Condition'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'Condition'

    key = 'dvdt_percentOfMax'
    theDict[key] = {}
    theDict[key]['defaultValue'] = 0.1
    theDict[key]['type'] = 'float'
    theDict[key]['allowNone'] = False
    theDict[key]['units'] = 'Percent'
    theDict[key]['humanName'] = 'dV/dt Percent of max'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'For dV/dt detection, the final TOP is when dV/dt drops to this percent from dV/dt AP peak'

    key = 'onlyPeaksAbove_mV'
    theDict[key] = {}
    theDict[key]['defaultValue'] = None
    theDict[key]['type'] = 'float'
    theDict[key]['allowNone'] = True
    theDict[key]['units'] = 'mV'
    theDict[key]['humanName'] = 'Accept Peaks Above (mV)'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'Only accept APs with peaks above this value (mV)'

    key = 'onlyPeaksBelow_mV'
    theDict[key] = {}
    theDict[key]['defaultValue'] = None
    theDict[key]['type'] = 'float'
    theDict[key]['allowNone'] = True
    theDict[key]['units'] = 'mV'
    theDict[key]['humanName'] = 'Accept Peaks Below (mV)'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'Only accept APs below this value (mV)'

    # TODO: get rid of this and replace with foot
    key = 'doBackupSpikeVm'
    theDict[key] = {}
    theDict[key]['defaultValue'] = False
    theDict[key]['type'] = 'boolean'
    theDict[key]['allowNone'] = False
    theDict[key]['units'] = 'Boolean'
    theDict[key]['humanName'] = 'Backup Vm Spikes'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'If true, APs detected with just mV will be backed up until Vm falls to xxx'

    key = 'refractory_ms'
    theDict[key] = {}
    theDict[key]['defaultValue'] = 170
    theDict[key]['type'] = 'float'
    theDict[key]['allowNone'] = False
    theDict[key]['units'] = 'ms'
    theDict[key]['humanName'] = 'Minimum AP interval (ms)'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'APs with interval (wrt previous AP) less than this will be removed'

    key = 'peakWindow_ms'
    theDict[key] = {}
    theDict[key]['defaultValue'] = 100
    theDict[key]['type'] = 'float'
    theDict[key]['allowNone'] = False
    theDict[key]['units'] = 'ms'
    theDict[key]['humanName'] = 'Peak Window (ms)'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'Window after TOP (ms) to seach for AP peak (mV)'

    key = 'dvdtPreWindow_ms'
    theDict[key] = {}
    theDict[key]['defaultValue'] = 10
    theDict[key]['type'] = 'float'
    theDict[key]['allowNone'] = False
    theDict[key]['units'] = 'ms'
    theDict[key]['humanName'] = 'dV/dt Pre Window (ms)'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'Window (ms) to search before each TOP for real threshold crossing'

    key = 'dvdtPostWindow_ms'
    theDict[key] = {}
    theDict[key]['defaultValue'] = 20
    theDict[key]['type'] = 'float'
    theDict[key]['allowNone'] = False
    theDict[key]['units'] = 'ms'
    theDict[key]['humanName'] = 'dV/dt Post Window (ms)'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'Window (ms) to search after each AP peak for minimum in dv/dt'

    key = 'mdp_ms'
    theDict[key] = {}
    theDict[key]['defaultValue'] = 250
    theDict[key]['type'] = 'float'
    theDict[key]['allowNone'] = False
    theDict[key]['units'] = 'ms'
    theDict[key]['humanName'] = 'Pre AP MDP window (ms)'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'Window (ms) before an AP to look for MDP'

    key = 'avgWindow_ms'
    theDict[key] = {}
    theDict[key]['defaultValue'] = 5
    theDict[key]['type'] = 'float'
    theDict[key]['allowNone'] = False
    theDict[key]['units'] = 'ms'
    theDict[key]['humanName'] = 'MDP averaging window (ms)'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'Window (ms) to calculate MDP (mV) as a mean rather than mV at single point for MDP'

    key = 'lowEddRate_warning'
    theDict[key] = {}
    theDict[key]['defaultValue'] = 8
    theDict[key]['type'] = 'float'
    theDict[key]['allowNone'] = False
    theDict[key]['units'] = 'EDD slope'
    theDict[key]['humanName'] = 'EDD slope warning'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'Generate warning when EED slope is lower than this value.'

    key = 'halfHeights'
    theDict[key] = {}
    theDict[key]['defaultValue'] = [10, 20, 50, 80, 90]
    theDict[key]['type'] = 'list' # list of number
    theDict[key]['allowNone'] = False
    theDict[key]['units'] = ''
    theDict[key]['humanName'] = 'AP Durations (%)'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'AP Durations as percent of AP height (AP Peak (mV) - TOP (mV))'

    key = 'halfWidthWindow_ms'
    theDict[key] = {}
    theDict[key]['defaultValue'] = 200
    theDict[key]['type'] = 'float'
    theDict[key]['allowNone'] = False
    theDict[key]['units'] = 'ms'
    theDict[key]['humanName'] = 'Half Width Window (ms)'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'Window (ms) after TOP to look for AP Durations'

    key = 'preSpikeClipWidth_ms'
    theDict[key] = {}
    theDict[key]['defaultValue'] = 200
    theDict[key]['type'] = 'float'
    theDict[key]['allowNone'] = False
    theDict[key]['units'] = 'ms'
    theDict[key]['humanName'] = 'Pre AP Clip Width (ms)'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'The pre duration of generated AP clips (Before AP)'

    key = 'postSpikeClipWidth_ms'
    theDict[key] = {}
    theDict[key]['defaultValue'] = 500
    theDict[key]['type'] = 'float'
    theDict[key]['allowNone'] = False
    theDict[key]['units'] = 'ms'
    theDict[key]['humanName'] = 'Post AP Clip Width (ms)'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'The post duration of generated AP clips (After AP)'

    key = 'medianFilter'
    theDict[key] = {}
    theDict[key]['defaultValue'] = 0
    theDict[key]['type'] = 'int'
    theDict[key]['allowNone'] = True # 0 is no median filter (see SavitzkyGolay_pnts)
    theDict[key]['units'] = 'points'
    theDict[key]['humanName'] = 'Median Filter Points'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'Number of points in median filter, must be odd, 0 for no filter'

    key = 'SavitzkyGolay_pnts'
    theDict[key] = {}
    theDict[key]['defaultValue'] = 5 #20211001 was 5
    theDict[key]['type'] = 'int'
    theDict[key]['allowNone'] = True # 0 is no filter
    theDict[key]['units'] = 'points'
    theDict[key]['humanName'] = 'SavitzkyGolay Points'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'Number of points in SavitzkyGolay filter, must be odd, 0 for no filter'

    key = 'SavitzkyGolay_poly'
    theDict[key] = {}
    theDict[key]['defaultValue'] = 2
    theDict[key]['type'] = 'int'
    theDict[key]['allowNone'] = False
    theDict[key]['units'] = ''
    theDict[key]['humanName'] = 'SavitzkyGolay Poly Deg'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'The degree of the polynomial for Savitzky-Golay filter'

    # key = 'dateAnalyzed'
    # theDict[key] = {}
    # theDict[key]['defaultValue'] = ''
    # theDict[key]['type'] = 'str'
    # theDict[key]['allowNone'] = False
    # theDict[key]['units'] = ''
    # theDict[key]['humanName'] = 'Date Analyzed'
    # theDict[key]['errors'] = ('')
    # theDict[key]['description'] = 'The date of analysis (yyyymmdd)'

    key = 'verbose'
    theDict[key] = {}
    theDict[key]['defaultValue'] = False
    theDict[key]['type'] = 'boolean'
    theDict[key]['allowNone'] = False
    theDict[key]['units'] = 'Boolean'
    theDict[key]['humanName'] = 'Verbose'
    theDict[key]['errors'] = ('')
    theDict[key]['description'] = 'Verbose Detection Reporting'

    # assign each detection param current value to it default value
    for k,v in theDict.items():
        defaultValue = theDict[k]['defaultValue']
        theDict[k]['currentValue'] = defaultValue

    '''
    if detectionPreset == bDetection.detectionPresets.default:
        # these are defaults from above
        pass
    elif detectionPreset == bDetection.detectionPresets.sanode:
        # these are defaults from above
        pass
    elif detectionPreset == bDetection.detectionPresets.ventricular:
        theDict['dvdtThreshold']['defaultValue'] = 100
        theDict['mvThreshold']['defaultValue'] = -20
        theDict['refractory_ms']['defaultValue'] = 200  # max freq of 5 Hz
        theDict['peakWindow_ms']['defaultValue'] = 100
        theDict['halfWidthWindow_ms']['defaultValue'] = 300
        theDict['preSpikeClipWidth_ms']['defaultValue'] = 200
        theDict['postSpikeClipWidth_ms']['defaultValue'] = 500
    elif detectionPreset == bDetection.detectionPresets.neuron:
        theDict['dvdtThreshold']['defaultValue'] = 20
        theDict['mvThreshold']['defaultValue'] = -40
        theDict['refractory_ms']['defaultValue'] = 4
        theDict['peakWindow_ms']['defaultValue'] = 5
        theDict['halfWidthWindow_ms']['defaultValue'] = 4
        theDict['dvdtPreWindow_ms']['defaultValue'] = 2
        theDict['dvdtPostWindow_ms']['defaultValue'] = 2
        theDict['preSpikeClipWidth_ms']['defaultValue'] = 2
        theDict['postSpikeClipWidth_ms']['defaultValue'] = 2
    elif detectionPreset == bDetection.detectionPresets.fastneuron:
        theDict['dvdtThreshold']['defaultValue'] = 20
        theDict['mvThreshold']['defaultValue'] = -40
        theDict['refractory_ms']['defaultValue'] = 3
        theDict['peakWindow_ms']['defaultValue'] = 2
        theDict['halfWidthWindow_ms']['defaultValue'] = 4
        theDict['dvdtPreWindow_ms']['defaultValue'] = 2
        theDict['dvdtPostWindow_ms']['defaultValue'] = 2
        theDict['preSpikeClipWidth_ms']['defaultValue'] = 2
        theDict['postSpikeClipWidth_ms']['defaultValue'] = 2
    elif detectionPreset == bDetection.detectionPresets.subthreshold:
        theDict['dvdtThreshold']['defaultValue'] = math.nan
        theDict['mvThreshold']['defaultValue'] = -20  # user specifies
        theDict['refractory_ms']['defaultValue'] = 100  # max freq is 10 Hz
        theDict['peakWindow_ms']['defaultValue'] = 50
        theDict['halfWidthWindow_ms']['defaultValue'] = 100
        theDict['preSpikeClipWidth_ms']['defaultValue'] = 100
        theDict['postSpikeClipWidth_ms']['defaultValue'] = 200
        theDict['onlyPeaksAbove_mV']['defaultValue'] = None
        theDict['onlyPeaksBelow_mV']['defaultValue'] = -20
        # todo: add onlyPeaksBelow_mV
    elif detectionPreset == bDetection.detectionPresets.caspikes:
        #theDict['detectionType']['defaultValue'] = sanpy.bDetection.detectionTypes.mv # ('dvdt', 'mv')
        theDict['dvdtThreshold']['defaultValue'] = math.nan #if None then detect only using mvThreshold
        theDict['mvThreshold']['defaultValue'] = 0.5
        #theDict['refractory_ms']['defaultValue'] = 200 #170 # reject spikes with instantaneous frequency
        #theDict['halfWidthWindow_ms']['defaultValue'] = 200 #was 20
    elif detectionPreset == bDetection.detectionPresets.cakymograph:
        theDict['detectionType']['defaultValue'] = sanpy.bDetection.detectionTypes['mv'].value
        theDict['dvdtThreshold']['defaultValue'] = math.nan #if None then detect only using mvThreshold
        theDict['mvThreshold']['defaultValue'] = 1.2
        theDict['peakWindow_ms']['defaultValue'] = 700
        theDict['halfWidthWindow_ms']['defaultValue'] = 800
        theDict['refractory_ms']['defaultValue'] = 500
        theDict['doBackupSpikeVm']['defaultValue'] = False
        # theDict['SavitzkyGolay_pnts']['defaultValue'] = 5
        theDict['preSpikeClipWidth_ms']['defaultValue'] = 200
        theDict['postSpikeClipWidth_ms']['defaultValue'] = 1000
    else:
        logger.error(f'Did not understand detection type "{detectionPreset}"')
        logger.error(f'    bDetection.detectionPresets.fastneuron: {bDetection.detectionPresets.fastneuron}')
        logger.error(f'    type(bDetection.detectionPresets.fastneuron): {type(bDetection.detectionPresets.fastneuron)}')
        logger.error(f'    detectionPreset == bDetection.detectionPresets.fastneuron: {detectionPreset == bDetection.detectionPresets.fastneuron}')
    '''

    # assign each detection param current value to it default value
    '''
    for k,v in theDict.items():
        defaultValue = theDict[k]['defaultValue']
        theDict[k]['currentValue'] = defaultValue
    '''

    return theDict.copy()

def printDocs():
    """Print out human readable detection parameters and convert to markdown table.

    Requires:
        pip install tabulate

    See: bAnalysisResults.printDocs()
    """
    logger.info('')
    
    import pandas as pd

    #detectionPreset = bDetection.detectionPresets.default  # detectionPresets_ is an enum class
    d = getDefaultDetection()
    dictList = []
    for k,v in d.items():
        parameter = k
        oneDict = {
            'Parameter': parameter,
            'Default Value': v['defaultValue'],
            'Units': v['units'],
            'Human Readable': v['humanName'],
            'Description': v['description'],
        }
        dictList.append(oneDict)
    #
    df = pd.DataFrame(dictList)
    
    # spit out markdown to copy/paste into mkdocs md file
    # REMEMBER: This requires `pip install tabulate`
    #outStr = df.to_markdown()
    # print(outStr)
    
    # save to csv for making a table for manuscript
    path = '/Users/cudmore/Desktop/sanpy-detection-params-20230316.csv'
    print('saving to:', path)
    df.to_csv(path, index=False)

'''
class detectionPresets():
    """A list of bDetection that are saved/loaded from json files.
    """
    def __init__(self):
        self._path = 'detectionPresets'  # folder to load from
        
        self.loadPresets()
    
    def loadPresets(self):
        """Load all presets in folder 'detectionPresets'
        """
        
        logger.info('')
        files = os.dir(self._path)
        for file in files:
            if not file.endswith('.json'):
                continue
            oneJson = json_load(file)
            print(oneJson)
'''

class bDetection(object):
    """ Class to manage detection parameters.
    """
    detectionTypes = detectionTypes_
    """ Enum with the type of spike detection, (dvdt, mv)"""

    def __init__(self):
        """Load all sanpy and <user> detection json files."""

        # dict with detection key and current value
        self._dDict = getDefaultDetection()
        
        # list of preset names including <user>SanPy/detection json files
        # use item=e[key] or item=e(value) then use item.name or item.value
        _theDict, _userPresetsDict = self._getPresetsDict()
        self._detectionEnum = Enum('self.detectionEnum', _theDict)

        # dictionary of presets, each key is like 'sanode' and then value is a dict of
        #   detection param keys and their values
        self._detectionPreset = {}
        for item in self._detectionEnum:
            # item is like self.detectionEnum.sanode
            #logger.info(f'  loaded {item}, {item.name}, "{item.value}"')
            presetValues = self._getPresetValues(item)
            if presetValues:
                # got value of built in preset
                self._detectionPreset[item.name] = presetValues
            else:
                # find in user loaded presets
                _userKey = item.name
                try:
                    self._detectionPreset[item.name] = _userPresetsDict[_userKey]
                except:
                    logger.erro(f'did not find item.name:{item.name} in _detectionPreset {self._detectionPreset.keys()}')

    def getDetectionPresetList(self):
        """"Get list of names of detection type.

        Used to make a list in popup in interface/ and interface/plugins
        """
        detectionList = []
        for detectionPreset in self._detectionEnum:
            detectionList.append(detectionPreset.value)
        return detectionList

    def _getPresetsDict(self):
        """Load detection presets from json files in 2 different folder:
            1) sanpy/detection-presets
            2) <user>/Documents/Sanpy/detection/

        For file name like 'Fast Neuron.json' make key 'fastneuron'

        Returns:
            userPresets (dict): One item per user file
        """
        
        def fileNameToKey(filePath):
            """Given full path to json file, return 
                Filename without extension, and a well formed key.
            """
            fileName = os.path.split(filePath)[1]
            fileName = os.path.splitext(fileName)[0]

            fileNameKey = os.path.split(filePath)[1]
            fileNameKey = os.path.splitext(fileNameKey)[0]
            # reduce preset json filename to (lower case, no spaces, no dash)
            fileNameKey = fileNameKey.lower()
            fileNameKey = fileNameKey.replace(' ', '')
            fileNameKey = fileNameKey.replace('-', '')  # in case user specifies a '-' in file name
            return fileName, fileNameKey

        theDict= {}
        userPresets = {}

        #
        # get files in our 'detection-presets' folder
        '''
        presetsPath = pathlib.Path(sanpy._util.getBundledDir()) / 'detection-presets' / '*.json'
        files = glob.glob(str(presetsPath))
        for filePath in files:
            fileName, fileNameKey = fileNameToKey(filePath)
            #
            theDict[fileNameKey] = fileName
        '''
        theDict['sanode'] = 'SA Node'
        theDict['ventricular'] = 'Ventricular'
        theDict['neuron'] = 'Neuron'
        theDict['fastneuron'] = 'Fast Neuron'
        theDict['subthreshold'] = 'Sub Threshold'
        theDict['caspikes'] = 'Ca Spikes'
        theDict['cakymograph'] = 'Ca Kymograph'

        #
        # get files in <users>/Documents/SanPy/detection/ folder
        #userDetectionPath = pathlib.Path(sanpy._util._getUserDetectionFolder()) / '*.json'
        #files = glob.glob(str(userDetectionPath))
        files = self._getUserFiles()
        #print('user files:', files)
        for filePath in files:
            fileName, fileNameKey = fileNameToKey(filePath)
            theDict[fileNameKey] = fileName

            # load user preset json and grab (param keys and values)
            with open(filePath, 'r') as f:
                userPresetsDict = json.load(f)
                userPresets[fileNameKey] = userPresetsDict

        return theDict, userPresets

    def _getUserFiles(self):
        """Get the full path to all user file presets .json
        """
        userDetectionPath = pathlib.Path(sanpy._util._getUserDetectionFolder())
        if userDetectionPath.is_dir:
            files = userDetectionPath.glob('*.json')
            return files
        else:
            return []

    def toJson(self):
        """Get key and defaultValue
        """
        theDict = {}
        for k,v in self._dDict.items():
            theDict[k] = v['defaultValue']

        #logger.info('theDict:')
        #pprint(theDict)

        #with open(savePath, 'w') as f:
        #    json.dump(self._dDict, f, indent=4)
        theJson = json.dumps(theDict, indent=4)
        #logger.info('theJson:')
        return theJson

    def old_saveAs(self, detectionType : str, filename, path=None):
        """Save a detection dictionary to json.
        
        If running in GUI, main SanPy app will specify the correct path.

        This is only for user defined sets, we never save our built in detection (hard coded in code).

        Args:
            detectionType: human readable like 'SA Node', will become 'SA Node.json'
            filename: Name of file to save (no extension, will append .json)
        """
        #_keyName = self._detectionEnum(detectionType).name
        if path is None:
            # get <user>/Documents/SanPy/xxx folder
            savePath = sanpy._util._getUserDetectionFolder()
            savePath = pathlib.Path(savePath) / f'{filename}.json'
        logger.info(str(savePath))
        with open(savePath, 'w') as f:
            dDict = self.getDetectionDict(detectionType)
            dDict['detectionName'] = filename
            json.dump(dDict, f, indent=4)

    def printDict(self):
        for k in self._dDict.keys():
            v = self._dDict[k]['currentValue']
            print(f'  {k}: "{v}" {type(v)}')

    def getMasterDict(self, detectionType : str):
        """Get the full dictionary from self._dDict getDefaultDetection()

        This is needed by sanpy/interface/plugins/detectionParams.py
        """

        retDict = copy.deepcopy(self._dDict)

        detectionTypeKey = self._detectionEnum(detectionType).name
        oneType = self._detectionPreset[detectionTypeKey]
        for k,v in oneType.items():
            retDict[k]['currentValue'] = v

        return retDict

    def getDetectionDict(self, detectionType : str, allParameter=True):
        """Get a full detection dict.
        
        Presets like 'SA Node' only over-ride a subset of the defaults, thus need to merge.
        User saved detection parameters will have all keys.
        
        Args:
            detectionType : detection type key, like 'SA Node'
        """
        try:
            if allParameter:
                dDict = {}
                
                # master template to get all default values
                for k,v in self._dDict.items():
                    dDict[k] = v['currentValue']

                # use specified detection type to get over-written values
                detectionTypeKey = self._detectionEnum(detectionType).name
                oneType = self._detectionPreset[detectionTypeKey]
                for k,v in oneType.items():
                    dDict[k] = v
            return dDict
        except (KeyError) as e:
            logger.error(f'Did not find detectionType:"{detectionType}"')
        except (ValueError) as e:
            logger.error(f'Did not find detectionType:"{detectionType}"')

    def getValue(self, detectionType : str, key):
        """Get current value from key. Valid keys are defined in getDefaultDetection().
        
        Args:
            detectionType : string value from enum, like 'SA Node'
        """
        try:
            #return self._dDict[key]['currentValue']
            return self._detectionPreset[detectionType][key]
        except (KeyError) as e:
            logger.warning(f'Did not find detectionType "{detectionType}" or key "{key}" to get current value')
            # TODO: define default when not found ???
            return None

    def setValue(self, detectionType : str, key : str, value):
        """
        Set current value for key. Valid keys are defined in getDefaultDetection.

        For float values that need to take on none, value comes in as -1e9
        
        Args:
            detectionType : short name (not value)
        """
        try:
            valueType = type(value)
            valueIsNumber = isinstance(value, numbers.Number)
            valueIsString = isinstance(value, str)
            valueIsBool = isinstance(value, bool)
            valueIsList = isinstance(value, list)
            valueIsNone = value is None

            # from the master list
            expectedType = self._dDict[key]['type'] # (number, string, boolean)
            allowNone = self._dDict[key]['allowNone'] # used to turn off a detection param

            #logger.info(f'expectedType:{expectedType} value type is {type(value)}')

            if allowNone and valueIsNone:
                pass
            elif expectedType=='number' and not valueIsNumber:
                logger.warning(f'Type mismatch (number) setting key "{key}", got {valueType}, expecting {expectedType}')
                return False
            elif expectedType=='string' and not valueIsString:
                logger.warning(f'Type mismatch (string) setting "{key}", got {valueType}, expecting {expectedType}')
                return False
            elif expectedType=='boolean' and not valueIsBool:
                logger.warning(f'Type mismatch (bool) setting "{key}", got {valueType}, expecting {expectedType}')
                return False
            elif expectedType=='list' and not valueIsList:
                logger.warning(f'Type mismatch (list) setting "{key}", got {valueType}, expecting {expectedType}')
                return False
            '''
            elif expectedType=='sanpy.bDetection.detectionTypes':
                try:
                    value = sanpy.bDetection.detectionTypes[value].name
                    #print(value == sanpy.bDetection.detectionTypes.dvdt)
                    #print(value == sanpy.bDetection.detectionTypes.mv)
                except (KeyError) as e:
                    logger.error(f'sanpy.bDetection.detectionTypes does not contain value "{value}"')
            '''
            #
            # set
            #self._dDict[key]['currentValue'] = value
            self._detectionPreset[detectionType][key] = value

            logger.info(f"now detectionType:{detectionType} key:{key}: {self._detectionPreset[detectionType][key]} {type(self._detectionPreset[detectionType][key])}")

            return True

        except (KeyError) as e:
            logger.warning(f'Did not find detectionType:{detectionType}, key:"{key}" to set current value to "{value}"')
            return False

    def old_save(self, saveBase):
        """
        Save underlying dict to json file

        Args:
            save base (str): basename to append '-detection.json'
        """

        # convert

        savePath = saveBase + '-detection.json'

        with open(savePath, 'w') as f:
            json.dump(self._dDict, f, indent=4)

    def old_load(self, loadBase):
        """
        Load detection from json file.

        Fill in underlying dict
        """

        loadPath = loadBase + '-detection.json'

        if not os.path.isfile(loadPath):
            logger.error(f'Did not find file: {loadPath}')
            return

        with open(loadPath, 'r') as f:
            self._dDict = json.load(f)

        # convert

    def _getPresetValues(self, detectionPreset):
        """
        detectionName : corresponds to key in enum self._detectionEnum
        """

        theDict = {}
        
        if detectionPreset == self._detectionEnum.sanode:
            theDict['detectionName'] = 'SA Node'
            theDict['dvdtThreshold'] = 20
            theDict['mvThreshold'] = -20
            theDict['refractory_ms'] = 170  # max freq of 5 Hz
            theDict['peakWindow_ms'] = 100
            theDict['halfWidthWindow_ms'] = 200
            theDict['preSpikeClipWidth_ms'] = 200
            theDict['postSpikeClipWidth_ms'] = 500
        elif detectionPreset == self._detectionEnum.ventricular:
            theDict['detectionName'] = 'Ventricular'
            theDict['dvdtThreshold'] = 100
            theDict['mvThreshold'] = -20
            theDict['refractory_ms'] = 200  # max freq of 5 Hz
            theDict['peakWindow_ms'] = 100
            theDict['halfWidthWindow_ms'] = 300
            theDict['preSpikeClipWidth_ms'] = 200
            theDict['postSpikeClipWidth_ms'] = 500
        elif detectionPreset == self._detectionEnum.neuron:
            theDict['detectionName'] = 'Neuron'
            theDict['dvdtThreshold'] = 20
            theDict['mvThreshold'] = -40
            theDict['refractory_ms'] = 4
            theDict['peakWindow_ms'] = 5
            theDict['halfWidthWindow_ms'] = 4
            theDict['dvdtPreWindow_ms'] = 2
            theDict['dvdtPostWindow_ms'] = 2
            theDict['preSpikeClipWidth_ms'] = 2
            theDict['postSpikeClipWidth_ms'] = 2
        elif detectionPreset == self._detectionEnum.fastneuron:
            theDict['detectionName'] = 'Fast Neuron'
            theDict['dvdtThreshold'] = 20
            theDict['mvThreshold'] = -40
            theDict['refractory_ms'] = 3
            theDict['peakWindow_ms'] = 2
            theDict['halfWidthWindow_ms'] = 4
            theDict['dvdtPreWindow_ms'] = 2
            theDict['dvdtPostWindow_ms'] = 2
            theDict['preSpikeClipWidth_ms'] = 2
            theDict['postSpikeClipWidth_ms'] = 2
        elif detectionPreset == self._detectionEnum.subthreshold:
            theDict['detectionName'] = 'Subthreshold'
            theDict['dvdtThreshold'] = math.nan
            theDict['mvThreshold'] = -20  # user specifies
            theDict['refractory_ms'] = 100  # max freq is 10 Hz
            theDict['peakWindow_ms'] = 50
            theDict['halfWidthWindow_ms'] = 100
            theDict['preSpikeClipWidth_ms'] = 100
            theDict['postSpikeClipWidth_ms'] = 200
            theDict['onlyPeaksAbove_mV'] = None
            theDict['onlyPeaksBelow_mV'] = -20
            # todo: add onlyPeaksBelow_mV
        elif detectionPreset == self._detectionEnum.caspikes:
            theDict['detectionName'] = 'Ca Spikes'
            #theDict['detectionType'] = sanpy.bDetection.detectionTypes.mv # ('dvdt', 'mv')
            theDict['dvdtThreshold'] = math.nan #if None then detect only using mvThreshold
            theDict['mvThreshold'] = 0.5
            #theDict['refractory_ms'] = 200 #170 # reject spikes with instantaneous frequency
            #theDict['halfWidthWindow_ms'] = 200 #was 20
        elif detectionPreset == self._detectionEnum.cakymograph:
            theDict['detectionName'] = 'Ca Kymograph'
            theDict['detectionType'] = sanpy.bDetection.detectionTypes['mv'].value
            theDict['dvdtThreshold'] = 1.2 # rosie, was math.nan #if None then detect only using mvThreshold
            theDict['mvThreshold'] = 1.2
            theDict['peakWindow_ms'] = 400 # rosie, was 700
            theDict['halfWidthWindow_ms'] = 400 # rosie, was 800
            theDict['refractory_ms'] = 500
            theDict['doBackupSpikeVm'] = False
            # theDict['SavitzkyGolay_pnts'] = 5
            theDict['preSpikeClipWidth_ms'] = 200
            theDict['postSpikeClipWidth_ms'] = 1000
        else:
            logger.error(f'did not understand detectionPreset: {detectionPreset}')

        return theDict

def test_0():
    """
    Testing get/set of detection params
    """

    bd = bDetection()

    #xxx = bd.getValue('xxx')

    #ok = bd.setValue('xxx', 2)
    #print('ok:', ok)

    ok1 = bd.setValue('dvdtThreshold', None)
    if not ok1:
        print('ok1:', ok1)

    ok1_5 = bd.setValue('mvThreshold', None)
    if not ok1_5:
        print('failure ok ok1_5:', ok1_5)

    ok2 = bd.setValue('cellType', '111')
    if not ok2:
        print('ok2:', ok2)

    # for setting list, check that (i) not empty and (ii) list[i] == expected type
    ok3 = bd.setValue('halfHeights', [])
    if not ok3:
        print('ok3:', ok3)

    # start/stop seconds defaults to None but we want 'number'
    ok4 = bd.setValue('startSeconds', 1e6)
    if not ok4:
        print('ok4:', ok4)

    tmpDict = {'Idx': 2.0, 'Include': 1.0, 'File': '19114001.abf', 'Dur(s)': 60.0, 'kHz': 20.0, 'Mode': 'fix', 'Cell Type': '', 'Sex': '', 'Condition': '', 'Start(s)': math.nan, 'Stop(s)': math.nan, 'dvdtThreshold': 50.0, 'mvThreshold': -20.0, 'refractory_ms': math.nan, 'peakWindow_ms': math.nan, 'halfWidthWindow_ms': math.nan, 'Notes': ''}
    for k,v in tmpDict.items():
        print('  ', k, ':', v)
    okSetFromDict = bd.setFromDict(tmpDict)

def test_save_load():
    # load an abf
    path = '/Users/cudmore/Sites/SanPy/data/19114001.abf'
    ba = sanpy.bAnalysis(path)

    # analyze
    ba.spikeDetect()
    print(ba)

    #print(ba._getSaveFolder())
    saveBase = ba._getSaveBase()

    detectionClass = ba.detectionClass

    '''
    detectionClass.save(saveBase)
    detectionClass._dDict = None

    detectionClass.load(saveBase)

    detectionClass.print()
    '''

    # might work
    ba.saveAnalysis()

    ba.loadAnalysis()

    #ba.detectionClass.print()

    print(ba)

def _exportDetectionJson():
    """20220705
    
    need to load/save detection presets from json

    To start, this will save json from the hard coded getDefaultDetection() function.
    """
    savePath = '/Users/cudmore/Desktop/detection-presets'
    
    for oneDetection in bDetection.detectionPresets:
        print('oneDetection:', oneDetection, 'oneDetection.value:', oneDetection.value)
 
        onePreset = bDetection.detectionPresets(oneDetection)
        oneDetectionClass = bDetection(onePreset)
        #oneJson = oneDetectionClass.toJson()
       
        _theDict = {}
        for k,v in oneDetectionClass._dDict.items():
            print(k)
            _theDict[k] = v['defaultValue']

        #print('_theDict.items():', _theDict.items())
        #pprint(_theDict)

        oneSaveFile = oneDetection.value + '.json'
        onseSavePath = os.path.join(savePath, oneSaveFile)
        print('    ', onseSavePath)
        
        with open(onseSavePath, 'w') as f:
            #json.dump(oneJson, f, ensure_ascii=False, indent=4)
            json.dump(_theDict, f, indent=4)

        # load to check
        with open(onseSavePath) as f:
            # need this to keep them in order ... 'object_pairs_hook=OrderedDict'
            data_loaded = json.load(f, object_pairs_hook=OrderedDict)
        pprint(data_loaded)

        #break

def testSwitchType():

    #bd = sanpy.bDetection() # default to bDetection.detectionDefaults.default
    bd = sanpy.bDetection()
    
    # names of all presets, both (built in, and user)
    presetList = bd.getDetectionPresetList()
    print('presetList:', presetList)

    # detection dict for one
    print('=== sanode')
    dDict = bd.getDetectionDict('SA Node')
    for k,v in dDict.items():
        print(f'    {k}:{v} {type(v)}')

    print('=== fastneuron2')
    dDict = bd.getDetectionDict('Fast Neuron 2')
    for k,v in dDict.items():
        print(f'    {k}:{v} {type(v)}')

    # detection dict for all
    '''
    for item in bd._detectionEnum:
        dDict = bd.getDetectionDict(item.name)
        print(f'=== {item.name} {item.value}')
        pprint(dDict)
    '''

    bd.setValue('sanode', 'dvdtThreshold', 12)

    print('=== after setValue 12 sanode')
    dDict = bd.getDetectionDict('SA Node')
    for k,v in dDict.items():
        print(f'    {k}:{v} {type(v)}')

    newValue = bd.getValue('sanode', 'dvdtThreshold')
    print('newValue:', newValue)

    # works
    # bd.saveAs('SA Node', filename='xxx yyy')
    bd.saveAs('SA Node', filename='SA Node 3')
    
if __name__ == '__main__':
    #test_0()

    # this works
    printDocs()

    #test_save_load()

    #_exportDetectionJson()

    #testSwitchType()
    # print('detectionPresets.default:', bDetection.detectionPresets.default)
    # print('detectionPresets.neuron:', bDetection.detectionPresets.neuron)
    # print('detectionPresets.neuron.value:', bDetection.detectionPresets.neuron.value)
    # print('getDetectionPresetList:', bDetection.getDetectionPresetList())


import os

import sanpy

# from sanpy.analysisPlot import bAnalysisPlot

from sanpy.fileloaders.fileLoader_abf import fileLoader_abf
from sanpy.fileloaders.fileLoader_csv import fileLoader_csv
from sanpy.fileloaders.fileLoader_tif import fileLoader_tif

import logging
from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__, level=logging.DEBUG)

def test_fileLoader_tif():
    #path = 'data/kymograph/rosie-kymograph.tif'
    path = os.path.join('data', 'kymograph', 'rosie-kymograph.tif')
    tifFile = fileLoader_tif(path)
    assert tifFile is not None

    logger.info(f'tifFile: {tifFile}')

    # test plot
    # ap = bAnalysisPlot(tifFile)
    # ap.plotRaw()

    # import matplotlib.pyplot as plt
    # plt.show()

def test_fileLoader_csv():
    # path = 'data/19114001.csv'
    # path = 'data/2021_07_20_0010.csv'
    path = os.path.join('data', '2021_07_20_0010.csv')

    csvFile = fileLoader_csv(path)
    assert csvFile is not None

    csvFile._getDerivative() # calcullate derivates

    _filteredDeriv = csvFile.filteredDeriv
    assert _filteredDeriv.shape == csvFile.sweepX.shape

    # logger.info(f'csvFile: {csvFile}')

    csvFile.setSweep(0)

    # logger.info(f'dataPointsPerMs: {csvFile.dataPointsPerMs}')
    # logger.info(f'sweepX: {csvFile.sweepX.shape}')
    # logger.info(f'sweepY: {csvFile.sweepY.shape}')
    # logger.info(f'sweepC: {csvFile.sweepC.shape}')

    # test plot
    # ap = bAnalysisPlot(csvFile)
    # ap.plotRaw()

    # import matplotlib.pyplot as plt
    # plt.show()

def test_fileLoader_abf():
    # path = 'data/19114001.abf'
    # path = 'data/2021_07_20_0010.abf'
    path = os.path.join('data', '2021_07_20_0010.abf')

    abfFile = fileLoader_abf(path)
    assert abfFile is not None

    # logger.info(f'abfFile: {abfFile}')

    assert abfFile.loadFileType == 'abf'

    assert abfFile.sweepLabelX == 'sec'

    abfFile._getDerivative() # calcullate derivates

    _filteredDeriv = abfFile.filteredDeriv
    assert _filteredDeriv.shape == abfFile.sweepX.shape

    # print(abfFile)

    # test plot
    # ap = bAnalysisPlot(abfFile)
    # ap.plotRaw()

    # import matplotlib.pyplot as plt
    # plt.show()

    # utility to export an abf as csv (use once)
    if 0:
        import pandas as pd
        df = pd.DataFrame()
        df['sec'] = abfFile.sweepX

        for _sweep in abfFile.sweepList:
            abfFile.setSweep(_sweep)
            _sweepY = abfFile.sweepY

            colStr = 'mv_' + str(_sweep)
            df[colStr] = abfFile.sweepY

            # DAC

        _folder, csvFileName = os.path.split(path)
        csvFileName = os.path.splitext(csvFileName)[0] + '.csv'
        csvPath = os.path.join(_folder, csvFileName)
        print('csvPath:', csvPath)
        df.to_csv(csvPath, index=False)

    
def test_new_b_analysis():
    # test new version of bAnalysis using fileLoader
    # path = 'data/19114001.abf'
    path = os.path.join('data', '19114001.abf')
    ba = sanpy.bAnalysis(path)

    # path = 'data/19114001.csv'
    path = os.path.join('data', '19114001.csv')
    ba = sanpy.bAnalysis(path)

if __name__ == '__main__':
    #test_fileLoader_abf()
    test_fileLoader_csv()


# Author: Robert H Cudmore
# Date: 20190722

import os, sys
import numpy as np
import scipy.signal
from PyQt5 import QtWidgets, QtGui, QtCore
import matplotlib
import matplotlib.figure
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
import matplotlib.pyplot as plt # abb 202012 added to set theme
import matplotlib.ticker as ticker

#import qdarkstyle

# abb 20200718
# needed to import from SanPy which is one folder up
#sys.path.append("..")

#from SanPy import bAnalysis
#from SanPy import bAnalysisPlot
#import bAnalysis
#import bAnalysisPlot
from sanpy import bAnalysis
#import sanpy.bAnalysis

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

class CustomStyle(QtWidgets.QProxyStyle):
    def styleHint(self, hint, option=None, widget=None, returnData=None):
        if hint == QtWidgets.QStyle.SH_SpinBox_KeyPressAutoRepeatRate:
            return 10**10
        elif hint == QtWidgets.QStyle.SH_SpinBox_ClickAutoRepeatRate:
            return 10**10
        elif hint == QtWidgets.QStyle.SH_SpinBox_ClickAutoRepeatThreshold:
            # You can use only this condition to avoid the auto-repeat,
            # but better safe than sorry ;-)
            return 10**10
        else:
            return super().styleHint(hint, option, widget, returnData)

class draggable_lines:
    def __init__(self, ax, xPos, yPos,
                    hLength=5, vLength=20,
                    linewidth=3, color='k', doPick=False):
        self.ax = ax
        self.c = ax.get_figure().canvas

        self.xPos = xPos
        self.yPos = yPos

        self.hLineLength = hLength
        self.vLineLength = vLength

        # horz line (user can drag this)
        x = [xPos, xPos+hLength]
        y = [yPos, yPos]
        self.hLine, = self.ax.plot(x, y, '-',
                                label='xScaleBar',
                                linewidth=linewidth, c=color,
                                clip_on = False,
                                picker=5)
        '''
        self.hLine = matplotlib.lines.Line2D(x, y, label='xScaleBar',
                                marker=None,
                                linewidth=linewidth, c=color,
                                clip_on = False,
                                picker=5)
        self.ax.add_line(self.hLine)
        '''

        # vert line
        x = [xPos, xPos]
        y = [yPos, yPos+vLength]
        self.vLine, = self.ax.plot(x, y, '-',
                                label='yScaleBar',
                                linewidth=linewidth, c=color,
                                clip_on = False)
                                #picker=None)
        '''
        self.vLine = matplotlib.lines.Line2D(x, y, label='yScaleBar',
                                marker=None,
                                linewidth=linewidth, c=color,
                                clip_on = False,
                                picker=None)
        self.ax.add_line(self.vLine)
        '''

        self.c.draw_idle()
        self.sid = self.c.mpl_connect('pick_event', self.clickonline)

    def clickonline(self, event):
        #print('clickonline()')
        if event.artist == self.hLine:
            #print("  line selected ", event.artist)
            self.follower = self.c.mpl_connect("motion_notify_event", self.followmouse)
            #self.releaser = self.c.mpl_connect("button_press_event", self.releaseonclick)
            self.releaser = self.c.mpl_connect("button_release_event", self.releaseonclick)

    def hideScaleBar(self, xChecked, yChecked):
        xPos = np.nan
        yPos = np.nan

        if xChecked:
            self.hLine.set_ydata([self.yPos, self.yPos])
            self.hLine.set_xdata([self.xPos, self.xPos + self.hLineLength])
        else:
            self.hLine.set_ydata([yPos, yPos])
            self.hLine.set_xdata([xPos, xPos + self.hLineLength])

        # a second line print('Vline is vertical')
        if yChecked:
            self.vLine.set_xdata([self.xPos, self.xPos])
            self.vLine.set_ydata([self.yPos, self.yPos + self.vLineLength])
        else:
            self.vLine.set_xdata([xPos, xPos])
            self.vLine.set_ydata([yPos, yPos + self.vLineLength])

        self.c.draw_idle()

    def setWidthHeight(self, width=None, height=None):
        """
        set the size (width/height) of a scale bar
        """
        if width is not None:
            self.hLineLength = width
        if height is not None:
            self.vLineLength = height

        # refresh
        self.setPos()

    def setThickness(self, thickness):
        self.hLine.set_linewidth(thickness)
        self.vLine.set_linewidth(thickness)

    def setPos(self, xPos=None, yPos=None, fromMax=False):
        """
        use this while (i) user-drag and (2) set x/y scale

        fromMax: USe this when setting from x/y axes max and we will shift position

        """
        #print('draggable_lines.setPos()', 'xPos:', xPos, 'yPos:', yPos, 'fromMax:', fromMax)

        if xPos is None:
            pass
            #xPos = self.xPos
        else:
            if fromMax:
                self.xPos = xPos - 1.5 * self.hLineLength
            else:
                self.xPos = xPos
        if yPos is None:
            pass
            #yPos = self.yPos
        else:
            if fromMax:
                self.yPos = yPos - 1.5 * self.vLineLength
            else:
                self.yPos = yPos

        xPos = self.xPos
        yPos = self.yPos

        #print('  xPos:', xPos, 'yPos:', yPos)

        self.hLine.set_ydata([yPos, yPos])
        self.hLine.set_xdata([xPos, xPos + self.hLineLength])
        # a second line print('Vline is vertical')
        self.vLine.set_xdata([xPos, xPos])
        self.vLine.set_ydata([yPos, yPos + self.vLineLength])

        self.c.draw_idle()

    def followmouse(self, event):
        #print('followmouse()')
        self.setPos(event.xdata, event.ydata)

    def releaseonclick(self, event):
        #print('releaseonclick()')
        self.c.mpl_disconnect(self.releaser)
        self.c.mpl_disconnect(self.follower)

class bExportWidget(QtWidgets.QWidget):
    """
    Open a window and display the raw Vm of a bAnalysis abf file
    """

    myCloseSignal = QtCore.Signal(object)

    def __init__(self,
                    sweepX, sweepY,
                    xyUnits=('',''),
                    path='',
                    darkTheme=False,
                    xMin=None,
                    xMax=None,
                    xMargin=0,
                    type='vm', # (vm, dvdt, meanclip)
                    parent=None,
                    ):
        """
        """
        super(bExportWidget, self).__init__(parent)

        self._inCallback = False

        self.shortcut = QtWidgets.QShortcut(QtGui.QKeySequence("Ctrl+w"), self)
        self.shortcut.activated.connect(self.myCloseAction)

        #self.setFont(QtGui.QFont("Helvetica", 11, QtGui.QFont.Normal, italic=False))

        self.myType = type # use this to size defaults for scale bar
        
        #self.myParent = parent

        self.mySweepX = sweepX
        self.mySweepY = sweepY
        self.mySweepX_Downsample = self.mySweepX
        self.mySweepY_Downsample = self.mySweepY

        #okGo = self.setFile(file)
        self.path = path
        self.xyUnits = xyUnits

        self.darkTheme = darkTheme

        if self.darkTheme:
            plt.style.use('dark_background')
        else:
            plt.rcParams.update(plt.rcParamsDefault)

        self.scaleBarDict = {}
        self.scaleBarDict['hLength'] = 5
        self.scaleBarDict['vLength'] = 20
        self.scaleBarDict['lineWidth'] = 5
        if self.darkTheme:
            self.scaleBarDict['color'] = 'w'
        else:
            self.scaleBarDict['color'] = 'k'
        if self.myType == 'vm':
            self.scaleBarDict['hLength'] = 1 # second
            self.scaleBarDict['vLength'] = 20 # mv
        elif self.myType == 'dvdt':
            self.scaleBarDict['hLength'] = 1 # second
            self.scaleBarDict['vLength'] = 10
        elif self.myType == 'meanclip':
            self.scaleBarDict['hLength'] = 5 # ms
            self.scaleBarDict['vLength'] = 20 # mv

        self.xMargin = xMargin # seconds
        self.internalUpdate = False

        self.initUI()

        if xMin is not None and xMax is not None:
            pass
            #self.xMin = xMin
            #self.xMax = xMax
        else:
            xMin = np.nanmin(self.mySweepX_Downsample)
            xMax = np.nanmax(self.mySweepX_Downsample)
        #
        self.myAxis.set_xlim(xMin, xMax)

        #self._setXAxis(xMin, xMax)

    def switchFile(self, ba : bAnalysis, x, y):
        
        self.mySweepX = x  #ba.fileLoader.sweepX
        self.mySweepY = y  #ba.fileLoader.sweepY
        self.mySweepX_Downsample = self.mySweepX
        self.mySweepY_Downsample = self.mySweepY

        self.path = ba.fileLoader.filepath

        if self.path:
            windowTitle = os.path.split(self.path)[1]
            #print('windowTitle:', windowTitle)
            self.setWindowTitle('Export Trace: ' + windowTitle)
        else:
            self.setWindowTitle('Export Trace: ' + 'None')

        self.plotRaw()

        # TODO: put in plot()
        xMin = np.nanmin(self.mySweepX_Downsample)
        xMax = np.nanmax(self.mySweepX_Downsample)
        logger.info(f'xMin:{xMin} xMax:{xMax}')
        self.myAxis.set_xlim(xMin, xMax)

        try:
            self.myAxis.relim()  # causing exception: ValueError: array of sample points is empty
        except (ValueError) as e:
            logger.warning(f'relim triggered ValueError?')

        self.myAxis.autoscale_view(True,True,True)
        #plt.draw()

    def myCloseAction(self):
        self.closeEvent()

    def closeEvent(self, event=None):
        """
        in Qt, close only hides the widget!
        """
        #print('bExportWidget.closeEvent()')
        self.deleteLater()
        self.myCloseSignal.emit(self)

    def initUI(self):

        #self.setStyleSheet(qdarkstyle.load_stylesheet(qt_api='pyqt5'))

        '''
        myPath = os.path.dirname(os.path.abspath(__file__))
        mystylesheet_css = os.path.join(myPath, 'css', 'mystylesheet.css')
        myStyleSheet = None
        if os.path.isfile(mystylesheet_css):
            with open(mystylesheet_css) as f:
                myStyleSheet = f.read()

        if myStyleSheet is not None:
            self.setStyleSheet(myStyleSheet)
        '''

        #self.setGeometry(100, 100, 1000, 600)
        self.center()

        if self.path:
            windowTitle = os.path.split(self.path)[1]
            self.setWindowTitle('Export Trace: ' + windowTitle)
        else:
            self.setWindowTitle('Export Trace: ' + 'None')

        myAlignLeft = QtCore.Qt.AlignLeft
        myAlignTop = QtCore.Qt.AlignTop

        hMasterLayout = QtWidgets.QHBoxLayout()
        hMasterLayout.setAlignment(QtCore.Qt.AlignTop)
        self.setLayout(hMasterLayout)

        left_container = QtWidgets.QWidget(self)
        left_container.setFixedWidth(350)

        hMasterLayout.addWidget(left_container, myAlignTop)

        vBoxLayout = QtWidgets.QVBoxLayout(left_container) # VBox for controls
        vBoxLayout.setAlignment(QtCore.Qt.AlignTop)

        hBoxRow0 = QtWidgets.QHBoxLayout()
        vBoxLayout.addLayout(hBoxRow0, myAlignTop)

        #
        # first row of controls

        # x axis on/off (todo: does not need self)
        self.xAxisCheckBox = QtWidgets.QCheckBox('')
        self.xAxisCheckBox.setToolTip('Toggle X-Axis On/Off')
        self.xAxisCheckBox.setChecked(True)
        self.xAxisCheckBox.stateChanged.connect(self.xAxisToggle)
        hBoxRow0.addWidget(self.xAxisCheckBox, myAlignLeft)

        # x min
        xMinLabel = QtWidgets.QLabel('X-Min')
        hBoxRow0.addWidget(xMinLabel, myAlignLeft)
        self.xMinSpinBox = QtWidgets.QDoubleSpinBox()
        self.xMinSpinBox.setToolTip('X-Axis Minimum')
        self.xMinSpinBox.setSingleStep(0.1)
        self.xMinSpinBox.setMinimum(-1e6)
        self.xMinSpinBox.setMaximum(1e6)
        self.xMinSpinBox.setValue(0)
        self.xMinSpinBox.setToolTip('X-Axis Minimum')
        self.xMinSpinBox.setKeyboardTracking(False)
        self.xMinSpinBox.valueChanged.connect(self._setXAxis)
        #self.lineWidthSpinBox.editingFinished.connect(self._setLineWidth)
        hBoxRow0.addWidget(self.xMinSpinBox, myAlignLeft)

        # x max
        xMax = np.nanmax(self.mySweepX_Downsample)  # self.mySweepX_Downsample[-1]
        xMaxLabel = QtWidgets.QLabel('X-Max')
        hBoxRow0.addWidget(xMaxLabel, myAlignLeft)
        self.xMaxSpinBox = QtWidgets.QDoubleSpinBox()
        self.xMaxSpinBox.setToolTip('X-Axis Maximum')
        self.xMaxSpinBox.setSingleStep(0.1)
        self.xMaxSpinBox.setMinimum(-1e6)
        self.xMaxSpinBox.setMaximum(1e6)
        self.xMaxSpinBox.setValue(xMax)
        self.xMaxSpinBox.setToolTip('X-Axis Maximum')
        self.xMaxSpinBox.setKeyboardTracking(False)
        self.xMaxSpinBox.valueChanged.connect(self._setXAxis)
        #self.lineWidthSpinBox.editingFinished.connect(self._setLineWidth)
        hBoxRow0.addWidget(self.xMaxSpinBox, myAlignLeft)

        #
        # second row
        hBoxRow1 = QtWidgets.QHBoxLayout()
        vBoxLayout.addLayout(hBoxRow1)

        # y axis
        self.yAxisCheckBox = QtWidgets.QCheckBox('')
        self.yAxisCheckBox.setToolTip('Toggle Y-Axis On/Off')
        self.yAxisCheckBox.setChecked(True)
        self.yAxisCheckBox.stateChanged.connect(self.yAxisToggle)
        hBoxRow1.addWidget(self.yAxisCheckBox)

        # y min
        yMinLabel = QtWidgets.QLabel('Y-Min')
        hBoxRow1.addWidget(yMinLabel)
        yMinValue = np.nanmin(self.mySweepY_Downsample)
        self.yMinSpinBox = QtWidgets.QDoubleSpinBox()
        self.yMinSpinBox.setSingleStep(0.1)
        self.yMinSpinBox.setMinimum(-1e6)
        self.yMinSpinBox.setMaximum(1e6)
        self.yMinSpinBox.setValue(yMinValue) # flipped
        self.yMinSpinBox.setToolTip('Y-Axis Minimum')
        self.yMinSpinBox.setKeyboardTracking(False)
        self.yMinSpinBox.valueChanged.connect(self._setYAxis)
        #self.lineWidthSpinBox.editingFinished.connect(self._setLineWidth)
        hBoxRow1.addWidget(self.yMinSpinBox)

        # y max
        yMaxLabel = QtWidgets.QLabel('Y-Max')
        hBoxRow1.addWidget(yMaxLabel)
        yMaxValue = np.nanmax(self.mySweepY_Downsample)
        self.yMaxSpinBox = QtWidgets.QDoubleSpinBox()
        self.yMaxSpinBox.setSingleStep(0.1)
        self.yMaxSpinBox.setMinimum(-1e6)
        self.yMaxSpinBox.setMaximum(1e6)
        self.yMaxSpinBox.setValue(yMaxValue) # flipped
        self.yMaxSpinBox.setToolTip('Y-Axis Maximum')
        self.yMaxSpinBox.setKeyboardTracking(False)
        self.yMaxSpinBox.valueChanged.connect(self._setYAxis)
        #self.lineWidthSpinBox.editingFinished.connect(self._setLineWidth)
        hBoxRow1.addWidget(self.yMaxSpinBox)

        #
        # one 1/2 row
        hBoxRow1_5 = QtWidgets.QHBoxLayout()
        vBoxLayout.addLayout(hBoxRow1_5)

        # x-tick major
        lineWidthLabel = QtWidgets.QLabel('X-Tick Major')
        hBoxRow1_5.addWidget(lineWidthLabel)
        self.xTickIntervalSpinBox = QtWidgets.QDoubleSpinBox()
        self.xTickIntervalSpinBox.setSingleStep(0.1)
        self.xTickIntervalSpinBox.setMinimum(0.0)
        self.xTickIntervalSpinBox.setMaximum(1e6)
        self.xTickIntervalSpinBox.setValue(10)
        self.xTickIntervalSpinBox.setToolTip('Major Tick Interval, 0 To Turn Off')
        self.xTickIntervalSpinBox.setKeyboardTracking(False)
        self.xTickIntervalSpinBox.valueChanged.connect(self._setTickMajorInterval)
        #self.lineWidthSpinBox.editingFinished.connect(self._setLineWidth)
        hBoxRow1_5.addWidget(self.xTickIntervalSpinBox)

        # x-tick minor
        lineWidthLabel = QtWidgets.QLabel('Minor')
        hBoxRow1_5.addWidget(lineWidthLabel)
        self.xTickMinorIntervalSpinBox = QtWidgets.QDoubleSpinBox()
        self.xTickMinorIntervalSpinBox.setSingleStep(0.1)
        self.xTickMinorIntervalSpinBox.setMinimum(0.0)
        self.xTickMinorIntervalSpinBox.setMaximum(1e6)
        self.xTickMinorIntervalSpinBox.setValue(10)
        self.xTickMinorIntervalSpinBox.setToolTip('Minor Tick Interval, 0 To Turn Off')
        self.xTickMinorIntervalSpinBox.setKeyboardTracking(False)
        self.xTickMinorIntervalSpinBox.valueChanged.connect(self._setTickMinorInterval)
        #self.lineWidthSpinBox.editingFinished.connect(self._setLineWidth)
        hBoxRow1_5.addWidget(self.xTickMinorIntervalSpinBox)

        #
        # one 3/4 row
        hBoxRow1_ytick = QtWidgets.QHBoxLayout()
        vBoxLayout.addLayout(hBoxRow1_ytick)

        # y-tick major
        lineWidthLabel = QtWidgets.QLabel('Y-Tick Major')
        hBoxRow1_ytick.addWidget(lineWidthLabel)
        self.yTickIntervalSpinBox = QtWidgets.QDoubleSpinBox()
        self.yTickIntervalSpinBox.setSingleStep(0.1)
        self.yTickIntervalSpinBox.setMinimum(0.0)
        self.yTickIntervalSpinBox.setMaximum(1e6)
        self.yTickIntervalSpinBox.setValue(20)
        self.yTickIntervalSpinBox.setToolTip('Major Y-Tick Interval, 0 To Turn Off')
        self.yTickIntervalSpinBox.setKeyboardTracking(False)
        self.yTickIntervalSpinBox.valueChanged.connect(self._setYTickMajorInterval)
        #todo: FIX THIS RECURSION WITH USING UP/DOWN ARROWS
        #self.yTickIntervalSpinBox.editingFinished.connect(self._setYTickMajorInterval)
        hBoxRow1_ytick.addWidget(self.yTickIntervalSpinBox)

        # y-tick minor
        lineWidthLabel = QtWidgets.QLabel('Minor')
        hBoxRow1_ytick.addWidget(lineWidthLabel)
        self.yTickMinorIntervalSpinBox = QtWidgets.QDoubleSpinBox()
        self.yTickMinorIntervalSpinBox.setSingleStep(0.1)
        self.yTickMinorIntervalSpinBox.setMinimum(0.0)
        self.yTickMinorIntervalSpinBox.setMaximum(1e6)
        self.yTickMinorIntervalSpinBox.setValue(20)
        self.yTickMinorIntervalSpinBox.setToolTip('Minor Y-Tick Interval, 0 To Turn Off')
        self.yTickMinorIntervalSpinBox.setKeyboardTracking(False)
        self.yTickMinorIntervalSpinBox.valueChanged.connect(self._setYTickMinorInterval)
        #self.lineWidthSpinBox.editingFinished.connect(self._setLineWidth)
        hBoxRow1_ytick.addWidget(self.yTickMinorIntervalSpinBox)

        #
        # third row
        hBoxRow2 = QtWidgets.QHBoxLayout()
        vBoxLayout.addLayout(hBoxRow2)

        # x margin
        xMaxLabel = QtWidgets.QLabel('X-Margin')
        hBoxRow2.addWidget(xMaxLabel)
        self.xMarginSpinBox = QtWidgets.QDoubleSpinBox()
        self.xMarginSpinBox.setToolTip('X-Axis Maximum')
        self.xMarginSpinBox.setSingleStep(0.1)
        self.xMarginSpinBox.setMinimum(0)
        self.xMarginSpinBox.setMaximum(1e6)
        self.xMarginSpinBox.setValue(self.xMargin)
        self.xMarginSpinBox.setKeyboardTracking(False)
        self.xMarginSpinBox.valueChanged.connect(self._setXMargin)
        #self.lineWidthSpinBox.editingFinished.connect(self._setLineWidth)
        hBoxRow2.addWidget(self.xMarginSpinBox)

        #
        # fourth row
        hBoxRow3 = QtWidgets.QHBoxLayout()
        vBoxLayout.addLayout(hBoxRow3)

        # line width
        lineWidthLabel = QtWidgets.QLabel('Line Width')
        hBoxRow3.addWidget(lineWidthLabel)
        self.lineWidthSpinBox = QtWidgets.QDoubleSpinBox()
        self.lineWidthSpinBox.setSingleStep(0.1)
        self.lineWidthSpinBox.setMinimum(0.01)
        self.lineWidthSpinBox.setMaximum(100.0)
        self.lineWidthSpinBox.setValue(0.5)
        self.lineWidthSpinBox.setKeyboardTracking(False)
        self.lineWidthSpinBox.valueChanged.connect(self._setLineWidth)
        #self.lineWidthSpinBox.editingFinished.connect(self._setLineWidth)
        hBoxRow3.addWidget(self.lineWidthSpinBox)


        # color
        colorList = ['red', 'green', 'blue', 'cyan', 'magenta', 'yellow', 'black', 'white']
        wIdx = colorList.index('white')
        kIdx = colorList.index('black')
        colorLabel = QtWidgets.QLabel('Line Color')
        hBoxRow3.addWidget(colorLabel)
        self.colorDropdown = QtWidgets.QComboBox()
        self.colorDropdown.addItems(colorList)
        self.colorDropdown.setCurrentIndex(wIdx if self.darkTheme else kIdx)
        self.colorDropdown.currentIndexChanged.connect(self._setLineColor)
        hBoxRow3.addWidget(self.colorDropdown)

        #
        # fifth row
        hBoxRow4 = QtWidgets.QHBoxLayout()
        vBoxLayout.addLayout(hBoxRow4)

        # downsample
        downsampleLabel = QtWidgets.QLabel('Downsample')
        hBoxRow4.addWidget(downsampleLabel)
        self.downSampleSpinBox = QtWidgets.QSpinBox()
        self.downSampleSpinBox.setSingleStep(1)
        self.downSampleSpinBox.setMinimum(1)
        self.downSampleSpinBox.setMaximum(200)
        self.downSampleSpinBox.setValue(1)
        self.downSampleSpinBox.setKeyboardTracking(False)
        self.downSampleSpinBox.valueChanged.connect(self._setDownSample)
        #self.downSampleSpinBox.editingFinished.connect(self._setDownSample)
        hBoxRow4.addWidget(self.downSampleSpinBox)

        # meadianFilter
        medianFilterLabel = QtWidgets.QLabel('Median Filter (points)')
        hBoxRow4.addWidget(medianFilterLabel)
        self.medianFilterSpinBox = QtWidgets.QSpinBox()
        #self.medianFilterSpinBox.setStyle(CustomStyle())
        self.medianFilterSpinBox.setSingleStep(2)
        self.medianFilterSpinBox.setMinimum(1)
        self.medianFilterSpinBox.setMaximum(1000)
        self.medianFilterSpinBox.setValue(1)
        self.medianFilterSpinBox.setKeyboardTracking(False)
        self.medianFilterSpinBox.valueChanged.connect(self._setDownSample)
        #self.medianFilterSpinBox.editingFinished.connect(self._setDownSample)
        hBoxRow4.addWidget(self.medianFilterSpinBox)

        #
        # fifth row
        hBoxRow4_5 = QtWidgets.QHBoxLayout()
        vBoxLayout.addLayout(hBoxRow4_5, myAlignTop)

        # dark theme
        self.darkThemeCheckBox = QtWidgets.QCheckBox('Dark Theme')
        self.darkThemeCheckBox.setChecked(self.darkTheme)
        self.darkThemeCheckBox.stateChanged.connect(self._changeTheme)
        hBoxRow4_5.addWidget(self.darkThemeCheckBox)

        #
        # sixth row
        scaleBarGroupBox = QtWidgets.QGroupBox('Scale Bar')
        scaleBarGroupBox.setAlignment(myAlignTop)
        vBoxLayout.addWidget(scaleBarGroupBox, myAlignTop)

        gridBoxScaleBar = QtWidgets.QGridLayout()
        scaleBarGroupBox.setLayout(gridBoxScaleBar)

        hLength = self.scaleBarDict['hLength']
        vLength = self.scaleBarDict['vLength']
        lineWidth = self.scaleBarDict['lineWidth']

        # scale bar width (length)
        scaleBarWidthLabel = QtWidgets.QLabel('Width')
        gridBoxScaleBar.addWidget(scaleBarWidthLabel, 0, 0)
        self.scaleBarWidthSpinBox = QtWidgets.QDoubleSpinBox()
        self.scaleBarWidthSpinBox.setToolTip('X Scale Bar Width (0 to remove)')
        self.scaleBarWidthSpinBox.setSingleStep(0.1)
        self.scaleBarWidthSpinBox.setMinimum(-1e6)
        self.scaleBarWidthSpinBox.setMaximum(1e6)
        self.scaleBarWidthSpinBox.setValue(hLength)
        self.scaleBarWidthSpinBox.setKeyboardTracking(False)
        self.scaleBarWidthSpinBox.valueChanged.connect(self._setScaleBarSize)
        #self.lineWidthSpinBox.editingFinished.connect(self._setLineWidth)
        gridBoxScaleBar.addWidget(self.scaleBarWidthSpinBox, 0, 1)

        # scale bar height (length)
        scaleBarHeightLabel = QtWidgets.QLabel('Height')
        gridBoxScaleBar.addWidget(scaleBarHeightLabel, 1, 0)
        self.scaleBarHeightSpinBox = QtWidgets.QDoubleSpinBox()
        self.scaleBarHeightSpinBox.setToolTip('Y Scale Bar Height (0 to remove)')
        self.scaleBarHeightSpinBox.setSingleStep(0.1)
        self.scaleBarHeightSpinBox.setMinimum(-1e6)
        self.scaleBarHeightSpinBox.setMaximum(1e6)
        self.scaleBarHeightSpinBox.setValue(vLength)
        self.scaleBarHeightSpinBox.setKeyboardTracking(False)
        self.scaleBarHeightSpinBox.valueChanged.connect(self._setScaleBarSize)
        #self.lineWidthSpinBox.editingFinished.connect(self._setLineWidth)
        gridBoxScaleBar.addWidget(self.scaleBarHeightSpinBox, 1, 1)

        # scale bar line thickness
        scaleBarThicknessLabel = QtWidgets.QLabel('Thickness')
        gridBoxScaleBar.addWidget(scaleBarThicknessLabel, 2, 0)
        self.scaleBarThicknessSpinBox = QtWidgets.QDoubleSpinBox()
        self.scaleBarThicknessSpinBox.setToolTip('Scale Bar Thickness')
        self.scaleBarThicknessSpinBox.setSingleStep(1)
        self.scaleBarThicknessSpinBox.setMinimum(0.1)
        self.scaleBarThicknessSpinBox.setMaximum(1e6)
        self.scaleBarThicknessSpinBox.setValue(lineWidth)
        self.scaleBarThicknessSpinBox.setKeyboardTracking(False)
        self.scaleBarThicknessSpinBox.valueChanged.connect(self._setScaleBarThickness)
        gridBoxScaleBar.addWidget(self.scaleBarThicknessSpinBox, 2, 1)

        # save button
        saveButton = QtWidgets.QPushButton('Save', self)
        saveButton.resize(saveButton.sizeHint())
        saveButton.clicked.connect(self.save)
        vBoxLayout.addWidget(saveButton)

        #vBoxLayout.addStretch()

        self.figure = matplotlib.figure.Figure()
        self.canvas = FigureCanvas(self.figure)

        # set defaullt save name
        baseName = 'export'
        if self.path:
            baseName = os.path.splitext(self.path)[0]
        self.canvas.get_default_filename = lambda: f'{baseName}'

        # matplotlib navigation toolbar
        self.toolbar = NavigationToolbar(self.canvas, self)
        #self.toolbar.zoom()

        # need self. here to set theme
        self.plotVBoxLayout = QtWidgets.QVBoxLayout()
        self.plotVBoxLayout.addWidget(self.toolbar)
        self.plotVBoxLayout.addWidget(self.canvas)

        hMasterLayout.addLayout(self.plotVBoxLayout) #, stretch=8)

        #self.myAxis = None
        self.plotRaw(firstPlot=True)

        self.show()

    def _changeTheme(self):
        """
        to change the theme, we need to redraw everything
        """
        checked = self.darkThemeCheckBox.isChecked()

        # get x/y axes limits
        xMin, xMax = self.myAxis.get_xlim()
        yMin, yMax = self.myAxis.get_ylim()

        # remove
        self.plotVBoxLayout.removeWidget(self.toolbar)
        self.plotVBoxLayout.removeWidget(self.canvas)

        self.toolbar.setParent(None)
        self.canvas.setParent(None)

        self.figure = None # ???
        self.toolbar = None
        self.canvas = None

        #self.canvas.draw()
        #self.repaint()

        # set theme
        if checked:
            plt.style.use('dark_background')
            self.darkTheme = True
        else:
            plt.rcParams.update(plt.rcParamsDefault)
            self.darkTheme = False

        self.figure = matplotlib.figure.Figure()
        self.canvas = FigureCanvas(self.figure)

        # matplotlib navigation toolbar
        self.toolbar = NavigationToolbar(self.canvas, self)
        #self.toolbar.zoom()

        self.plotVBoxLayout.addWidget(self.toolbar)
        self.plotVBoxLayout.addWidget(self.canvas)
        # self.grid.addWidget(self.toolbar, ??)

        if self.darkTheme:
            self.scaleBarDict['color'] = 'w'
        else:
            self.scaleBarDict['color'] = 'k'

        #self.myAxis = None
        self.plotRaw(firstPlot=True)

        # restore original x/y axes
        self.myAxis.set_xlim(xMin, xMax)
        self.myAxis.set_ylim(yMin, yMax)

    def _setXMargin(self):
        self.xMarginSpinBox.setEnabled(False)
        self.xMargin = self.xMarginSpinBox.value()

        self.plotRaw()
        self.xMarginSpinBox.setEnabled(True)

    def _setXAxis(self):
        """
        called when user sets values in x spin boxes
        see: on_xlim_change
        """
        self.xMinSpinBox.setEnabled(False)
        self.xMaxSpinBox.setEnabled(False)

        xMin = self.xMinSpinBox.value()
        xMax = self.xMaxSpinBox.value()

        #print('_setXAxis() calling self.myAxis.set_xlim()', xMin, xMax)
        xMin -= self.xMargin
        xMax += self.xMargin

        self.internalUpdate = True
        self.myAxis.set_xlim(xMin, xMax)
        self.internalUpdate = False

        self.scaleBars.setPos(xPos=xMax, fromMax=True)

        #self.plotRaw(xMin=xMin, xMax=xMax)
        self.plotRaw()

        self.xMinSpinBox.setEnabled(True)
        self.xMaxSpinBox.setEnabled(True)

    def _setYAxis(self):
        """
        called when user sets values in y spin boxes
        see: on_xlim_change
        """
        self.yMinSpinBox.setEnabled(False)
        self.yMaxSpinBox.setEnabled(False)

        yMin = self.yMinSpinBox.value()
        yMax = self.yMaxSpinBox.value()

        self.myAxis.set_ylim(yMin, yMax)

        self.scaleBars.setPos(yPos=yMax, fromMax=True)

        self.yMinSpinBox.setEnabled(True)
        self.yMaxSpinBox.setEnabled(True)

    def on_xlims_change(self, mplEvent):
        """
        matplotlib callback
        """

        if self.internalUpdate:
            return

        xLim = mplEvent.get_xlim()
        #print('on_xlims_change() xLim:', xLim)

        xMin = xLim[0]
        xMax = xLim[1]

        self.xMinSpinBox.setValue(xMin)
        self.xMaxSpinBox.setValue(xMax)

        self.plotRaw()

        #self.canvas.draw_idle()

    def _setDownSample(self):
        self.downSampleSpinBox.setEnabled(False)
        self.medianFilterSpinBox.setEnabled(False)

        try:
            downSample = self.downSampleSpinBox.value()
            medianFilter = self.medianFilterSpinBox.value()

            if (medianFilter % 2) == 0:
                medianFilter += 1

            #print('_setDownSample() downSample:', downSample, 'medianFilter:', medianFilter)

            #print('downSample ... please wait')
            self.mySweepX_Downsample = self.mySweepX[::downSample]
            self.mySweepY_Downsample = self.mySweepY[::downSample]

            if medianFilter > 1:
                #print('medianFilter ... please wait')
                self.mySweepY_Downsample = scipy.signal.medfilt(self.mySweepY_Downsample,
                                            kernel_size=medianFilter)

            #
            self.plotRaw()

            # refresh scale-bar
            xPos = np.nanmax(self.mySweepX_Downsample)  # self.mySweepX_Downsample[-1]
            yPos = np.nanmax(self.mySweepY_Downsample)
            #self.scaleBars.setPos(xPos, yPos, fromMax=True)

        except (Exception) as e:
            logger.exceptiion('EXCEPTION in _setDownSample():', e)

        self.canvas.draw_idle()
        #self.repaint()

        self.downSampleSpinBox.setEnabled(True)
        self.medianFilterSpinBox.setEnabled(True)

    def xAxisToggle(self):
        checked = self.xAxisCheckBox.isChecked()
        self._toggleAxis('bottom', checked)

    def yAxisToggle(self):
        checked = self.yAxisCheckBox.isChecked()
        self._toggleAxis('left', checked)

    def scaleBarToggle(self):
        xChecked = self.xScaleBarCheckBox.isChecked()
        yChecked = self.yScaleBarCheckBox.isChecked()
        self._toggleScaleBar(xChecked, yChecked)

    def _setScaleBarSize(self):
        self.scaleBarWidthSpinBox.setEnabled(False)
        self.scaleBarHeightSpinBox.setEnabled(False)

        width = self.scaleBarWidthSpinBox.value()
        height = self.scaleBarHeightSpinBox.value()
        self.scaleBars.setWidthHeight(width=width, height=height)
        #
        self.canvas.draw_idle()
        #self.repaint()

        self.scaleBarWidthSpinBox.setEnabled(True)
        self.scaleBarHeightSpinBox.setEnabled(True)

    def _setScaleBarThickness(self):
        self.scaleBarThicknessSpinBox.setEnabled(False)

        thickness = self.scaleBarThicknessSpinBox.value()
        self.scaleBars.setThickness(thickness)
        #
        self.canvas.draw_idle()
        #self.repaint()

        self.scaleBarThicknessSpinBox.setEnabled(True)

    def _setYTickMajorInterval(self): #, interval):
        self.yTickIntervalSpinBox.setEnabled(False)
        interval = self.yTickIntervalSpinBox.value()
        if interval == 0:
            ml = ticker.NullLocator()
        else:
            ml = ticker.MultipleLocator(interval)
        self.myAxis.yaxis.set_major_locator(ml)
        #
        self.canvas.draw_idle()
        #self.repaint()
        self.yTickIntervalSpinBox.setEnabled(True)

    def _setYTickMinorInterval(self, interval):
        self.yTickMinorIntervalSpinBox.setEnabled(False)
        if interval == 0:
            ml = ticker.NullLocator()
        else:
            ml = ticker.MultipleLocator(interval)
        self.myAxis.yaxis.set_minor_locator(ml)
        #
        self.canvas.draw_idle()
        #self.repaint()
        self.yTickMinorIntervalSpinBox.setEnabled(True)

    def _setTickMajorInterval(self, interval):
        self.xTickIntervalSpinBox.setEnabled(False)
        if interval == 0:
            ml = ticker.NullLocator()
        else:
            ml = ticker.MultipleLocator(interval)
        self.myAxis.xaxis.set_major_locator(ml)
        #
        self.canvas.draw_idle()
        #self.repaint()
        self.xTickIntervalSpinBox.setEnabled(True)

    def _setTickMinorInterval(self, interval):
        self.xTickMinorIntervalSpinBox.setEnabled(False)
        if interval == 0:
            ml = ticker.NullLocator()
        else:
            ml = ticker.MultipleLocator(interval)
        self.myAxis.xaxis.set_minor_locator(ml)
        #
        self.canvas.draw_idle()
        #self.repaint()
        self.xTickMinorIntervalSpinBox.setEnabled(True)

    def _toggleAxis(self, leftBottom, onOff):
        if leftBottom == 'bottom':
            self.myAxis.get_xaxis().set_visible(onOff)
            self.myAxis.spines['bottom'].set_visible(onOff)
        elif leftBottom == 'left':
            self.myAxis.get_yaxis().set_visible(onOff)
            self.myAxis.spines['left'].set_visible(onOff)
        #
        self.canvas.draw_idle()
        #self.repaint()

    def _toggleScaleBar(self, xChecked, yChecked):
        self.scaleBars.hideScaleBar(xChecked, yChecked)

    def _setLineColor(self, colorStr):

        colorStr = self.colorDropdown.currentText()

        self.myTraceLine.set_color(colorStr)
        '''
        for line in self.myAxis.lines:
            print('_setLineColor:', line.get_label())
            line.set_color(colorStr)
        '''

        #
        self.canvas.draw_idle()
        #self.repaint()

    def _setLineWidth(self):
        self.lineWidthSpinBox.setEnabled(False)

        lineWidth = self.lineWidthSpinBox.value()

        #print('bExportWidget._setLineWidth() lineWidth:', lineWidth)

        self.myTraceLine.set_linewidth(lineWidth)
        '''
        for line in self.myAxis.lines:
            line.set_linewidth(lineWidth)
        '''
        #
        self.canvas.draw_idle()
        #self.repaint()

        self.lineWidthSpinBox.setEnabled(True)

    def plotRaw(self, xMin=None, xMax=None, firstPlot=False):
        if firstPlot:
            self.figure.clf()

            self.myAxis = self.figure.add_subplot(111)
            '''
            left = .2 #0.05
            bottom = 0.05
            width = 0.7 #0.9
            height = 0.9
            self.myAxis = self.figure.add_axes([left, bottom, width, height])
            '''

            self.myAxis.spines['right'].set_visible(False)
            self.myAxis.spines['top'].set_visible(False)

            if self.darkTheme:
                color = 'w'
            else:
                color = 'k'

            lineWidth = self.lineWidthSpinBox.value()

        sweepX = self.mySweepX_Downsample
        sweepY = self.mySweepY_Downsample

        if firstPlot:
            xMinOrig = sweepX[0]
            xMaxOrig = np.nanmax(sweepX)  # sweepX[-1]
        else:
            xMinOrig, xMaxOrig = self.myAxis.get_xlim()

        yMinOrig = np.nanmin(sweepY)
        yMaxOrig = np.nanmax(sweepY)

        xClip = self.xMargin
        if xMin is not None and xMax is not None:
            minClip = xMin + xClip
            maxClip = xMax - xClip
        else:
            minClip = xMinOrig + xClip
            maxClip = xMaxOrig - xClip

        sweepX = np.ma.masked_where( (sweepX<minClip), sweepX)
        sweepY = np.ma.masked_where( (sweepX<minClip), sweepY)

        sweepX = np.ma.masked_where( (sweepX>maxClip), sweepX)
        sweepY = np.ma.masked_where( (sweepX>maxClip), sweepY)

        #print('sweepX:', sweepX.shape, type(sweepX))
        #print('sweepY:', sweepY.shape, type(sweepY))

        if firstPlot:
            # using label 'myTrace' to differentiate from x/y scale bar
            self.myTraceLine, = self.myAxis.plot(sweepX, sweepY,
                                    '-', # fmt = '[marker][line][color]'
                                    c=color, linewidth=lineWidth,
                                    label='myTrace')
            #matplotlib.lines.Line2D

            self.myAxis.callbacks.connect('xlim_changed', self.on_xlims_change)

            # scale bar
            hLength = self.scaleBarDict['hLength']
            vLength = self.scaleBarDict['vLength']
            scaleBarLineWidth = self.scaleBarDict['lineWidth']
            scaleBarColor = self.scaleBarDict['color']
            xPos = xMaxOrig #sweepX[-1]
            yPos = yMaxOrig #np.nanmax(sweepY)
            self.scaleBars = draggable_lines(self.myAxis, xPos, yPos,
                                hLength=hLength, vLength=vLength,
                                linewidth=scaleBarLineWidth,
                                color=scaleBarColor,
                                doPick=True)
            self.scaleBars.setPos(xPos, yPos, fromMax=True)

        else:
            self.myTraceLine.set_xdata(sweepX)
            self.myTraceLine.set_ydata(sweepY)
            '''
            for line in self.myAxis.lines:
                print('plotRaw() is updating with set_xdata/set_ydata')
                line.set_xdata(sweepX)
                line.set_ydata(sweepY)
            '''

        #self.myAxis.use_sticky_edges = False
        #self.myAxis.margins(self.xMargin, tight=None)

        if firstPlot:
            #self.myAxis.set_ylabel('Vm (mV)')
            #self.myAxis.set_xlabel('Time (sec)')
            self.myAxis.set_ylabel(self.xyUnits[1])
            self.myAxis.set_xlabel(self.xyUnits[0])

        self.canvas.draw_idle()
        #self.repaint()

    def save(self):
        """
        Save the current view to a pdf file
        """

        # get min/max of x-axis
        [xMin, xMax] = self.myAxis.get_xlim()
        #if xMin < 0:
        #    xMin = 0

        xMin += self.xMargin
        xMax -= self.xMargin

        xMin = '%.2f'%(xMin)
        xMax = '%.2f'%(xMax)


        lhs, rhs = xMin.split('.')
        xMin = 'b' + lhs + '_' + rhs

        lhs, rhs = xMax.split('.')
        xMax = 'e' + lhs + '_' + rhs

        # construct a default save file name
        saveFilePath = ''
        if self.path:
            parentPath, filename = os.path.split(self.path)
            baseFilename, file_extension = os.path.splitext(filename)
            #saveFileName = baseFilename + '_' + self.myType + '_' + xMin + '_' + xMax + '.svg'
            saveFileName = f'{baseFilename}_{self.myType}_{xMin}_{xMax}.svg'
            #saveFileName = baseFilename + '.svg'
            saveFilePath = os.path.join(parentPath,saveFileName)

        # file save dialog
        #fullSavePath, ignore = QtWidgets.QFileDialog.getSaveFileName(self, 'Save File', saveFilePath, "pdf Files (*.pdf)")
        fullSavePath, ignore = QtWidgets.QFileDialog.getSaveFileName(self, 'Save File', saveFilePath)

        # do actual save
        if len(fullSavePath) > 0:
            logger.info(f'saving: {fullSavePath}')
            self.figure.savefig(fullSavePath)

    def center(self):
        """
        Center the window on the screen
        """
        qr = self.frameGeometry()
        cp = QtWidgets.QDesktopWidget().availableGeometry().center()
        qr.moveCenter(cp)
        self.move(qr.topLeft())

def run():
    path = '../data/19114001.abf'

    ba = bAnalysis(path)
    if ba.loadError:
        logger.error(f'Error loading file: {path}')
    else:
        app = QtWidgets.QApplication(sys.argv)
        app.aboutToQuit.connect(app.deleteLater)

        sweepX = ba.fileLoader.sweepX
        sweepY = ba.fileLoader.sweepY
        xyUnits = ('Time (sec)', 'Vm (mV)')
        type = 'vm'
        GUI = bExportWidget(sweepX, sweepY, path=path, xyUnits=xyUnits, type=type)
        GUI.show()

        sys.exit(app.exec_())

if __name__ == '__main__':
    run()


#20210619

from PyQt5 import QtCore, QtWidgets, QtGui

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

import sanpy
from sanpy.interface.plugins import sanpyPlugin
from sanpy.interface.plugins import ResponseType

class resultsTable2(sanpyPlugin):
    """
    Plugin to display 'human readable' summary of all spikes, one spike per row.

    Uses:
        sanpy.bExport.report2()
        QTableView: sanpy.interface.bErrorTable.errorTableView()
        QAbstractTableModel: sanpy.interface.bFileTable.pandasModel
    """
    myHumanName = 'Summary Spikes (Full)'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.toggleResponseOptions(ResponseType.setSweep, False)  # we plot all sweeps
        self.toggleResponseOptions(ResponseType.setAxis, False)

        self.cardiacDf = None

        layout = QtWidgets.QVBoxLayout()

        controlsLayout = QtWidgets.QHBoxLayout()
        self.numSpikesLabel = QtWidgets.QLabel('unknown spikes')
        controlsLayout.addWidget(self.numSpikesLabel)
        layout.addLayout(controlsLayout)

        self.myErrorTable = sanpy.interface.bErrorTable.errorTableView()
        # TODO: derive a more general purpose table, here we are re-using error table
        self.myErrorTable.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Interactive);
        self.myErrorTable.horizontalHeader().setStretchLastSection(False)

        layout.addWidget(self.myErrorTable)

        #self.setLayout(layout)
        self.getVBoxLayout().addLayout(layout)
        
        #
        # connect clicks in error table to signal main sanpy_app with slot_selectSpike()
        logger.warning('mar 11 FIX SPIKE SELECTION')
        if self.getSanPyApp() is not None:
            fnPtr = self.getSanPyApp().slot_selectSpike
            self.myErrorTable.signalSelectSpike.connect(fnPtr)

        self.replot()

    def setAxis(self):
        # inherited, resopnd to user setting x-axis
        self.replot()

    def replot(self):
        """Generate report with sanpy.bExport.report2()
        and put it into a table.
        """
        logger.info('')
        # update
        if self.ba is None:
            return

        myExport = sanpy.bExport(self.ba)

        startSec, stopSec = self.getStartStop()
        if startSec is None or stopSec is None:
            startSec = 0
            stopSec = self.ba.fileLoader.recordingDur

        self.cardiacDf = myExport.report2(startSec, stopSec) # report2 is more 'cardiac'

        logger.info(f' generated df len:{len(self.cardiacDf)} startSec:{startSec} stopSec:{stopSec}')

        errorReportModel = sanpy.interface.bFileTable.pandasModel(self.cardiacDf)
        self.myErrorTable.setModel(errorReportModel)

        for colIdx in range(len(self.cardiacDf.columns)):
            if colIdx == 0:
                continue
            self.myErrorTable.setColumnWidth(colIdx, 170)

        self.numSpikesLabel.setText(f'{len(self.cardiacDf)} spikes')

    def copyToClipboard(self):
        if self.ba is not None and self.cardiacDf is not None:
            logger.info('Copy to clipboard')
            self.cardiacDf.to_clipboard(sep='\t', index=False)

    def selectSpikeList(self):
        spikeList = self.getSelectedSpikes()
        logger.info(f'{spikeList}')
        self.myErrorTable.mySelectRows(spikeList)

def main():
    path = '/home/cudmore/Sites/SanPy/data/19114001.abf'
    ba = sanpy.bAnalysis(path)
    ba.spikeDetect()
    print(ba.numSpikes)

    import sys
    app = QtWidgets.QApplication([])
    rt = resultsTable2(ba=ba)
    rt.show()
    sys.exit(app.exec_())

if __name__ == '__main__':
    main()



from pprint import pprint

from PyQt5 import QtCore, QtWidgets, QtGui

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

import sanpy
from sanpy.interface.plugins import sanpyPlugin
from sanpy.interface.plugins import ResponseType

class resultsTable(sanpyPlugin):
    """Plugin to display summary of all spikes, one spike per row.

    Uses:
        QTableView: sanpy.interface.bErrorTable.errorTableView()
        QAbstractTableModel: sanpy.interface.bFileTable.pandasModel
    """
    myHumanName = 'Summary Spikes'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.toggleResponseOptions(ResponseType.setSweep, False)  # we plot all sweeps
        self.toggleResponseOptions(ResponseType.setAxis, False)

        #self.pyqtWindow() # makes self.mainWidget

        # layout = QtWidgets.QVBoxLayout()

        # topToolbarLayout0, topToolbarLayout1 = self._buildTopToolbar()  # horizontal toolbar with checkboxes to toggle signals
        # layout.addLayout(topToolbarLayout0)
        # layout.addLayout(topToolbarLayout1)

        # TODO: derive a more general purpose table, here we are re-using error table
        self.myErrorTable = sanpy.interface.bErrorTable.errorTableView()
        self.myErrorTable.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Interactive);
        self.myErrorTable.horizontalHeader().setStretchLastSection(False)

        #layout.addWidget(self.myErrorTable)
        self.getVBoxLayout().addWidget(self.myErrorTable)

        #self.setLayout(layout)

        #
        # connect clicks in error table to signal main sanpy_app with slot_selectSpike()
        logger.warning('FIX SPIKE SELECTION')
        if self.getSanPyApp() is not None:
            fnPtr = self.getSanPyApp().slot_selectSpike
            self.myErrorTable.signalSelectSpike.connect(fnPtr)

        self.replot()

    def setAxis(self):
        # inherited, resopnd to user setting x-axis
        self.replot()

    def _getReport(self):
        dfPlot = self.ba.asDataFrame()
        return dfPlot

    def replot(self):
        logger.info('')
        # update
        if self.ba is None:
            return

        startSec, stopSec = self.getStartStop()

        dfPlot = self._getReport()

        if dfPlot is not None:
            if startSec is not None and stopSec is not None:
                # use column thresholdSec
                dfPlot = dfPlot[ (dfPlot['thresholdSec']>=startSec) & (dfPlot['thresholdSec']<=stopSec)]
                #pass
            #
            logger.info(f'dfReportForScatter {startSec} {stopSec}')
            errorReportModel = sanpy.interface.bFileTable.pandasModel(dfPlot)
            self.myErrorTable.setModel(errorReportModel)

            # mar 21, when did i get rid of this?
            # self._fileLabel.setText(f'{self.ba.fileLoader.filename}')
            # self._numSpikesLabel.setText(f'{len(dfPlot)} spikes')

    def copyToClipboard(self):
        if self.ba is not None:
            dfPlot = self._getReport()
            if dfPlot is not None:
                logger.info('Copy to clipboard')
                dfPlot.to_clipboard(sep='\t', index=False)

def main():
    path = 'data/19114001.abf'
    ba = sanpy.bAnalysis(path)

    bd = sanpy.bDetection()  # gets default
    dDict = bd.getDetectionDict('SA Node')

    ba.spikeDetect(dDict)
    #print(ba.numSpikes)

    #print(ba.spikeDict._myList[0])
    #_df = ba.asDataFrame()
    #pprint(_df)

    # does not work, _myList[3] is class sanpy.analysisResults.analysisResult
    # print('qqqqqqqqqq')
    # print(ba.spikeDict._myList[3])

    #pprint(ba.spikeDict.asDataFrame())

    # print('=== 2')
    # print(ba.spikeDict[2])

    #sys.exit()

    import sys
    app = QtWidgets.QApplication([])
    rt = resultsTable(ba=ba)
    rt.show()
    sys.exit(app.exec_())

if __name__ == '__main__':
    main()



import numpy as np
import pandas as pd

from sanpy.fileloaders.fileLoader_base import fileLoader_base, recordingModes

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

class fileLoader_csv(fileLoader_base):
    loadFileType = 'csv'
    
    # @property
    # def loadFileType(self):
    #     return 'csv'
    
    def loadFile(self):
        """Load file and call setLoadedData().
        
        Use self.filepath for the file
       
        Column 0 is seconds
        Column 1 is recording in ['mV', 'pA']
        """
        # load the csv file
        _df = pd.read_csv(self.filepath)
        _columns = _df.columns
        _numColumns = len(_columns)
        if _numColumns < 2:
             logger.error(f'expecting 2 or more columns but got {_numColumns}')

        _numSamples = len(_df)
        _numSweeps = _numColumns - 1  # assuming no DAC columns

        # TODO: check that the columns we will read are float (not str or object)
        #           
        # make sweepX (assuming all sweeps have the same sweepX)
        sweepX = np.ndarray( (_numSamples,1))
        try:
            sweepX[:,0] = _df[_columns[0]].to_numpy()
        except (ValueError) as e:
            raise ValueError
        
        # make sweepY to hold (samples, sweeps) values, the values in each sweep are different
        sweepY = np.ndarray( (_numSamples,_numSweeps))
        for sweep in range(_numSweeps):
            colIdx = sweep + 1
            sweepY[:,sweep] = _df[_columns[colIdx]].to_numpy()
                
        #sweepC = None  # not defined

        # determine recording mode from name of column 1
        _tmpColName = _columns[1].replace('_0','')
        if _tmpColName in ['pA', 'nA']:
            _recordingMode = recordingModes.vclamp
        elif _tmpColName in ['mV', 'mv']:
            _recordingMode = recordingModes.iclamp
        else:
            logger.warning(f'did not infer recording mode from column name {_tmpColName}')
            _recordingMode = recordingModes.unknown

        xLabel = _columns[0]
        yLabel = _tmpColName  #_columns[1]
        
        self.setLoadedData(
                sweepX = sweepX,
                sweepY = sweepY,
                #sweepC = sweepC,
                xLabel = xLabel,
                yLabel = yLabel,
        )